unit AudioCaching;

interface

uses

  System.Classes, System.SysUtils, System.Generics.Collections,

  KnobsUtils, wave_tools, knobs2013;


const

  MAX_SAMPLE_COUNT = 128 * 1024 * 1024;
  AUTO_NORMALIZE   = True;


type

  TAudioRefList = class( TStringList)
  // Case sensitive, sorted StringList without duplicates and no empty strings
  public
    constructor Create;
    function    Add( const aValue: string): Integer;                                                           override;
  end;


  TAudioWaveInfo = class
  private
    FReader   : TWaveFileReader;
    FData     : TWaveData;
    FDataSize : UInt64;
    FRefCount : Integer;
  protected
    function    GetChannelCount: UInt16;
    function    GetSampleCount : UInt64;
    procedure   IncRefCount;
    procedure   DecRefCount;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    function    ReadData( const aRiffReader: TRiffReaderWriter; const aFileName: string): Boolean;
  protected
    property    DataSize     : UInt64          read FDataSize;
    property    RefCount     : Integer         read FRefCount       write FRefCount;
  public
    property    ChannelCount : UInt16          read GetChannelCount;
    property    SampleCount  : UInt64          read GetSampleCount;
    property    Reader       : TWaveFileReader read FReader;
    property    Data         : TWaveData       read FData;
  end;


  TAudioCache = class( TDictionary< string, TAudioWaveInfo>)
  private
    FRiffReader     : TRiffReaderWriter;
    FMaxSampleCount : Cardinal;
    FAdded          : Boolean;
    FErrorStr       : string;
    FDataSize       : UInt64;
    FUppedCount     : Integer;
  private
    function    GetUppedCount: Boolean;
    function    AddFile( const aFileName: string): Boolean;
  protected
    procedure   ValueNotify( const aValue: TAudioWaveInfo; anAction: TCollectionNotification);                 override;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    procedure   FreeWaveInfo( const aFileName: string);
    function    AddWaveInfo ( const aFileName: string): TAudioWaveInfo;
    procedure   IncRefCountsFor( const aFileNames: TAudioRefList);
    procedure   DecRefCountsFor( const aFileNames: TAudioRefList);
  public
    property    UpCount        : Boolean  read GetUppedCount;
    property    ErrorStr       : string   read FErrorStr;
    property    DataSize       : UInt64   read FDataSize;
    property    MaxSampleCount : Cardinal read FMaxSampleCount write FMaxSampleCount;
  end;


  function  GetAudioCache     : TAudioCache;
  function  GetAudioCacheSize : UInt64;
  function  CreateAudioRefList: TAudioRefList;
  procedure IncAudioRefCountsFor( const aFileNames: TAudioRefList);
  procedure DecAudioRefCountsFor( const aFileNames: TAudioRefList);


implementation


var

  GAudioCache : TAudioCache = nil;

  function  GetAudioCache: TAudioCache;
  begin
    if   not Assigned( GAudioCache)
    then GAudioCache := TAudioCache.Create;

    Result := GAudioCache;
  end;


  function  GetAudioCacheSize: UInt64;
  begin
    if   Assigned( GAudioCache)
    then Result := GAudioCache.DataSize
    else Result := 0;
  end;


  function  CreateAudioRefList: TAudioRefList;
  begin
    Result := TAudioRefList.Create;
  end;


  procedure IncAudioRefCountsFor( const aFileNames: TAudioRefList);
  begin
    if   Assigned( aFileNames )
    and  Assigned( GAudioCache)
    then GAudioCache.IncRefCountsFor( aFileNames);
  end;


  procedure DecAudioRefCountsFor( const aFileNames: TAudioRefList);
  begin
    if Assigned( aFilenames)
    then begin
      if   Assigned( GAudioCache)
      then GAudioCache.DecRefCountsFor( aFileNames);

      aFileNames.Clear;
    end;
  end;


  procedure CleanupUnit;
  begin
    if   Assigned( GAudioCache)
    then begin
      if   GAudioCache.DataSize <> 0
      then Nop;                              // Breakpoint here .. DataSize should be zero, or something was not freed

      FreeAndNil( GAudioCache);
    end;
  end;


{ ========
  TAudioRefList = class( TStringList)// Case sensitive, sorted StringList without duplicates and no empty strings
  public
}

    constructor TAudioRefList.Create;
    begin
      inherited;
      Duplicates    := dupAccept;
      CaseSensitive := True;
      Sorted        := True;
      OwnsObjects   := False;
    end;


    function    TAudioRefList.Add( const aValue: string): Integer; // override;
    begin
      if   aValue <> ''           // Don't add empty strings
      then Result := inherited
      else Result := -1;
    end;


{ ========
 TAudioWaveInfo = class
  private
    FReader   : TWaveFileReader;
    FData     : TWaveData;
    FDataSize : UInt64;
    FRefCount : Integer;
  protected
    property    DataSize     : UInt64          read FDataSize;
    property    RefCount     : Integer         read FRefCount       write FRefCount;
  public
    property    ChannelCount : UInt16          read GetChannelCount;
    property    SampleCount  : UInt64          read GetSampleCount;
    property    Reader       : TWaveFileReader read FReader;
    property    Data         : TWaveData       read FData;
  protected
}

    function    TAudioWaveInfo.GetChannelCount: UInt16;
    begin
      if   Assigned( FData)
      then Result := Length( FData)
      else Result := 0;;
    end;


    function    TAudioWaveInfo.GetSampleCount : UInt64;
    begin
      if   ChannelCount > 0
      then Result := Length( FData[ 0])
      else Result := 0;
    end;


    procedure   TAudioWaveInfo.IncRefCount;
    begin
      Inc( FRefCount);
    end;


    procedure   TAudioWaveInfo.DecRefCount;
    begin
      DecToZero( FRefCount);
    end;


//  public

    constructor TAudioWaveInfo.Create;
    begin
      SetLength( FData, 0);
      FRefCount             := 0;
      FDataSize             := 0;
      FReader               := TWaveFileReader.Create;
      FReader.AutoNormalize := AUTO_NORMALIZE;
    end;


    destructor  TAudioWaveInfo.Destroy; // override;
    begin
      FreeAndNil( FReader );

      SetLength( FData, 0);
      inherited;
    end;


    function    TAudioWaveInfo.ReadData( const aRiffReader: TRiffReaderWriter; const aFileName: string): Boolean;
    begin
       if   Assigned( aRiffReader)
       then begin
         if   aRiffReader.LoadFromFile( aFileName, MAX_SAMPLE_COUNT)
         then begin
           Result := FReader.ReadData( aRiffReader, FData);

           if   Result
           then begin
             FDataSize := SizeOf( FData);

             if   Length( FData) > 0
             then FDataSize := FDataSize * UInt64( Length( FData[ 0]));
           end;
         end
         else Result := False;
       end
       else Result := False;
    end;


 { ========
  TAudioCache = class( TDictionary< string, TAudioWaveInfo>)
  private
    FRiffReader     : TRiffReaderWriter;
    FMaxSampleCount : Cardinal;
    FAdded          : Boolean;
    FErrorStr       : string;
    FDataSize       : UInt64;
    FUppedCount     : Integer;
  public
    property    UppedCount     : Boolean  read GetUppedCount;
    property    ErrorStr       : string   read FErrorStr;
    property    DataSize       : UInt64   read FDataSize;
    property    MaxSampleCount : Cardinal read FMaxSampleCount write FMaxSampleCount;
  private
}

    function    TAudioCache.GetUppedCount: Boolean;
    begin
      Result := FUppedCount > 0;
    end;


    function    TAudioCache.AddFile( const aFileName: string): Boolean;
    var
      anInfo : TAudioWaveInfo;
    begin
      Result := False;

      if   Assigned( FRiffReader)
      and  ( aFileName <> '')
      and  FileExists( aFileName)
      then begin
        try
          FAdded := False;
          anInfo := TAudioWaveInfo.Create;

          if   not anInfo.ReadData( FRiffReader, aFileName)
          then begin
            if   anInfo.Reader.ErrorStr = ''
            then FErrorStr :=                                 FRiffReader.ErrorStr
            else FErrorStr := anInfo.Reader.ErrorStr + ' -' + FRiffReader.ErrorStr;
          end
          else Add( aFileName, anInfo);
        finally
          if   not FAdded
          then FreeAndNil( anInfo)
          else Result := True;
        end;
      end;
    end;


//  protected

    procedure   TAudioCache.ValueNotify( const aValue: TAudioWaveInfo; anAction: TCollectionNotification); // override;
    begin
      inherited;

      if   anAction = cnRemoved
      then begin
        FDataSize := FDataSize - aValue.DataSize;
        aValue.DisposeOf;
      end
      else if anAction = cnAdded
      then begin
        FAdded := True;
        FDataSize := FDataSize + aValue.DataSize;
      end;
    end;


// public

    constructor TAudioCache.Create;
    begin
      inherited Create;
      FRiffReader := TRiffReaderWriter.Create;
      FErrorStr   := '';
    end;


    destructor  TAudioCache.Destroy; // override;
    begin
      FreeAndNil( FRiffReader);
      Clear;
      inherited;
    end;


    procedure   TAudioCache.FreeWaveInfo( const aFileName: string);
    var
      anInfo: TAudioWaveInfo;
    begin
      if   aFileName <> ''
      then begin
        if   TryGetValue( aFileName, anInfo)
        then begin
          anInfo.DecRefCount;

          if   anInfo.RefCount <= 0
          then Remove( aFileName);
        end;
      end;
    end;


    function    TAudioCache.AddWaveInfo( const aFileName: string): TAudioWaveInfo;
    begin
      FErrorStr := '';

      if   aFileName <> ''
      then begin
        if   not TryGetValue( aFileName, Result)
        then begin
          if   AddFile( aFileName)
          then begin
            if   TryGetValue( aFileName, Result)
            then Result.IncRefCount;
          end
          else Result := nil;
        end
        else Result.IncRefCount;

        Nop;
      end
      else TryGetValue( aFileName, Result);
    end;


    procedure   TAudioCache.IncRefCountsFor( const aFileNames: TAudioRefList);
    var
      i      : Integer;
      anInfo : TAudioWaveInfo;
    begin
      if   Assigned( aFileNames)
      then begin
        for i := 0 to aFileNames.Count - 1
        do begin
          if   TryGetValue( aFileNames[ i], anInfo)
          then anInfo.IncRefCount;
        end;

        Inc( FUppedCount);
      end;
    end;


    procedure   TAudioCache.DecRefCountsFor( const aFileNames: TAudioRefList);
    var
      i : Integer;
    begin
      if   Assigned( aFileNames)
      and  ( FUppedCount > 0)
      then begin
        Dec( FUppedCount);

        for i := 0 to aFileNames.Count - 1
        do  FreeWaveInfo( aFileNames[ i]);
      end;
    end;


initialization


finalization


  CleanupUnit;


end.
