 unit Lockhart;

{
  Code found at : https://github.com/danmohd/LockhartWavefolderJUCE - no coypyright info available.

    This is an implementation of 'Virtual Analog Model of the Lockhart Wavefolder' Conference Paper 2017 by
    Fabian Esqueda, Julian D. Parker, Henry Pintynen and Stefan Bilbao.

  Translation to Pascal for use in Wren :

  // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

   COPYRIGHT 2021 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}

interface

uses

  System.SysUtils, System.Math,

  KnobsUtils, anti_derivatives;


type

  TLockharts = class
  private
    FLocked    : Boolean;
    FLockharts : TArray<TLockhart>;
    FR         : TArray<TSignal>;
    FGain      : TSignal;
    FOffset    : TSignal;
  private
    function    GetR( anIndex: Integer): TSignal;
    procedure   SetR( anIndex: Integer; aValue: TSignal);
  public
    constructor Create( aLength: Integer);
    destructor  Destroy;                                                                                      override;
    function    Process( const anInValue: TSignal): TSignal;
  public
    property    R[ anIndex: Integer] : TSignal read GetR    write SetR;
    property    Gain                 : TSignal read FGain   write FGain;
    property    Offset               : TSignal read FOffset write FOffset;
  end;


  TPolyLockharts = class
  private
    FPolyphony : Integer;
    FLength    : Integer;
    FLockharts : TArray<TLockharts>;
    FLocked    : Boolean;
  private
    function    GetOffset( aVoice: Integer): TSignal;
    procedure   SetOffset( aVoice: Integer; aValue: TSignal);
    function    GetGain  ( aVoice: Integer): TSignal;
    procedure   SetGain  ( aVoice: Integer; aValue: TSignal);
    function    GetR     ( aVoice, anIndex: Integer): TSignal;
    procedure   SetR     ( aVoice, anIndex: Integer; aValue: TSignal);
  public
    constructor Create( aPolyphony, aLength: Integer);
    destructor  Destroy;                                                                                      override;
    function    Process( aVoice: Integer; const anInValue: TSignal): TSignal;
  public
    property    R     [ aVoice, anIndex: Integer] : TSignal read GetR      write SetR;
    property    Gain  [ aVoice         : Integer] : TSignal read GetGain   write SetGain;
    property    Offset[ aVoice         : Integer] : TSignal read GetOffset write SetOffset;
  end;


implementation


{ ======
  TLockharts = class
  private
    FLocked    : Boolean;
    FLockharts : TArray<TLockhart>;
    FR         : TArray<TSignal>;
    FGain      : TSignal;
    FOffset    : TSignal;
  public
    property    R[ anIndex: Integer] : TSignal read GetR    write SetR;
    property    Gain                 : TSignal read FGain   write FGain;
    property    Offset               : TSignal read FOffset write FOffset;
  private
}

    function    TLockharts.GetR( anIndex: Integer): TSignal;
    begin
      Assert(( anIndex >= 0) and ( anIndex < Length( FR)));
      Result := FR[ anIndex];
    end;


    procedure   TLockharts.SetR( anIndex: Integer; aValue: TSignal);
    begin
      Assert(( anIndex >= 0) and ( anIndex < Length( FR)));
      FR[ anIndex] := aValue;
    end;


//   public

    constructor TLockharts.Create( aLength: Integer);
    var
      i : Integer;
    begin
      FLocked := True;
      inherited Create;
      SetLength( FLockharts, aLength);
      SetLength( FR        , aLength);

      for i := 0 to Length( FLockharts) - 1
      do  begin
        FR        [ i] := 5e3;
        FLockHarts[ i] := TLockhart.Create;
      end;

      Gain    := 3.0;
      FLocked := False;
    end;


    destructor  TLockharts.Destroy; // override;
    var
      i : Integer;
    begin
      FLocked := True;

      for i := 0 to Length( FLockharts) - 1
      do  FreeAndNil( FLockHarts[ i]);

      inherited;
    end;


    function    TLockharts.Process( const anInValue: TSignal): TSignal;
    var
      i : Integer;
    begin
      Result := anInValue * Gain + Offset;

      if   not FLocked
      then begin
        for i := 0 to Length( FLockharts) - 1
        do begin
          FLockharts[ i].RL := R[ i];
          Result := Clip( FLockharts[ i].Process( Result), -1e100, 1e100);
        end;
      end;

      Result := Result * 2.0;
    end;


{ ========
  TPolyLockharts = class
  private
    FPolyphony : Integer;
    FLength    : Integer;
    FLockharts : TArray<TLockharts>;
    FLocked    : Boolean;
  public
    property    R     [ aVoice, anIndex: Integer] : TSignal read GetR      write SetR;
    property    Gain  [ aVoice         : Integer] : TSignal read GetGain   write SetGain;
    property    Offset[ aVoice         : Integer] : TSignal read GetOffset write SetOffset;
  private
}

    function    TPolyLockharts.GetOffset( aVoice: Integer): TSignal;
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));
      Result := FLockharts[ aVoice].Offset;
    end;


    procedure   TPolyLockharts.SetOffset( aVoice: Integer; aValue: TSignal);
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));
      FLockharts[ aVoice].Offset := aValue;
    end;


    function    TPolyLockharts.GetGain( aVoice: Integer): TSignal;
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));
      Result := FLockharts[ aVoice].Gain;
    end;


    procedure   TPolyLockharts.SetGain( aVoice: Integer; aValue: TSignal);
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));
      FLockharts[ aVoice].Gain := aValue;
    end;


    function    TPolyLockharts.GetR( aVoice, anIndex: Integer): TSignal;
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));
      Result := FLockharts[ aVoice].R[ anIndex];
    end;


    procedure   TPolyLockharts.SetR( aVoice, anIndex: Integer; aValue: TSignal);
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));
      FLockharts[ aVoice].R[ anIndex] := aValue;
    end;


//   public

    constructor TPolyLockharts.Create( aPolyphony, aLength: Integer);
    var
      i : Integer;
      j : Integer;
    begin
      FLocked := True;
      inherited Create;
      SetLength( FLockharts, aPolyphony);

      for i := 0 to aPolyphony - 1
      do FLockharts[ i] := TLockharts.Create( aLength);

      FLength    := aLength;
      FPolyphony := aPolyphony;

      for i := 0 to aPolyphony - 1
      do begin
        FLockharts[ i].Gain   := 3.0;
        FLockharts[ i].Offset := 0.0;

        for j := 0 to FLength - 1
        do FLockharts[ i].R[ j] := 5e3;
      end;

      FLocked := False;
    end;


    destructor  TPolyLockharts.Destroy; // override;
    var
      i : Integer;
    begin
      FLocked := True;

      for i := 0 to FPolyphony - 1
      do FreeAndNil( FLockharts[ i]);

      inherited;
    end;


    function    TPolyLockharts.Process( aVoice: Integer; const anInValue: TSignal): TSignal;
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));

      if   not FLocked
      then Result := FLockharts[ aVoice].Process( anInValue)
      else Result := 0.0;
    end;


end.
