unit SVFUnit;

{

   COPYRIGHT 2016 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl


  original code :

    Original source : http://www.musicdsp.org/showone.php?id=142

    no copyright notice was found for that code.
}

interface

uses

  System.SysUtils, System.Math,

  KnobsUtils, KnobsConversions, FormantControl;


type

  TSvfMode = (
    svfmLowPass,
    svfmBandPass,
    svfmBandReject,
    svfmHighPass
  );


  TSvfChamberlin = class
  private
    type TState = record
      FInternalQ    : TSignal;
      FQ            : TSignal;
      FInternalFreq : TSignal;
      FFrequency    : TSignal;
      FDelay1       : TSignal;
      FDelay2       : TSignal;
    end;
    PState = ^ TState;
  private
    FPolyphony    : Integer;
    FSampleRate   : TSignal;
    FReciprocalSR : TSignal;
    FState        : array of TState;
  private
    procedure   FixFrequency( aVoice: Integer);
    procedure   FixQ        ( aVoice: Integer);
  private
    procedure   SetSampleRate( const aValue : TSignal);
    function    GetFrequency ( aVoice: Integer): TSignal;
    procedure   SetFrequency ( aVoice: Integer; const aValue : TSignal);
    function    GetQ         ( aVoice: Integer): TSignal;
    procedure   SetQ         ( aVoice: Integer; const aValue : TSignal);
  public
    constructor Create( aPolyphony: Integer);
    procedure   Process( aVoice: Integer; const anInput : TSignal; var aLowPass, aBandPass, aBandReject, aHighPass : TSignal);
  public
    property    Polyphony                    : Integer read FPolyphony;
    property    SampleRate                   : TSignal read FSampleRate  write SetSampleRate;
    property    Frequency [ aVoice: Integer] : TSignal read GetFrequency write SetFrequency;
    property    Q         [ aVoice: Integer] : TSignal read GetQ         write SetQ;
  end;


  TSvfChamberlinRelative = class
  private
    type TState = record
      FInternalQ    : TSignal;
      FQ            : TSignal;
      FInternalFreq : TSignal;
      FFrequency    : TSignal;    // in Units
      FLFM          : TSignal;
      FDelay1       : TSignal;
      FDelay2       : TSignal;
      FRelFrequency : TSignal;    // in Units
      FRelQ         : TSignal;
      FMainQ        : TSignal;
    end;
    PState = ^ TState;
  private
  FPolyphony      : Integer;
    FSampleRate   : TSignal;
    FReciprocalSR : TSignal;
    FState        : array of TState;
  private
    procedure   FixFrequency( aVoice: Integer);
    procedure   FixQ        ( aVoice: Integer);
  private
    procedure   SetSampleRate  ( const aValue : TSignal);
    function    GetFrequency   ( aVoice: Integer): TSignal;
    procedure   SetFrequency   ( aVoice: Integer; const aValue : TSignal);
    function    GetLFM         ( aVoice: Integer): TSignal;
    procedure   SetLFM         ( aVoice: Integer; const aValue : TSignal);
    function    GetQ           ( aVoice: Integer): TSignal;
    procedure   SetQ           ( aVoice: Integer; const aValue : TSignal);
    function    GetRelFrequency( aVoice: Integer): TSignal;
    procedure   SetRelFrequency( aVoice: Integer; const aValue : TSignal);
    function    GetRelQ        ( aVoice: Integer): TSignal;
    procedure   SetRelQ        ( aVoice: Integer; const aValue : TSignal);
    function    GetMainQ       ( aVoice: Integer): TSignal;
    procedure   SetMainQ       ( aVoice: Integer; const aValue : TSignal);
  public
    constructor Create( aPolyphony: Integer);
    procedure   Process( aVoice : Integer; const anInput : TSignal; var aLowPass, aBandPass, aBandReject, aHighPass : TSignal);
  public
    property    Polyphony                      : Integer read FPolyphony;
    property    SampleRate                     : TSignal read FSampleRate     write SetSampleRate;
    property    Frequency   [aVoice : Integer] : TSignal read GetFrequency    write SetFrequency;
    property    LFM         [aVoice : Integer] : TSignal read GetLFM          write SetLFM;
    property    Q           [aVoice : Integer] : TSignal read GetQ            write SetQ;
    property    RelFrequency[aVoice : Integer] : TSignal read GetRelFrequency write SetRelFrequency;
    property    RelQ        [aVoice : Integer] : TSignal read GetRelQ         write SetRelQ;
    property    MainQ       [aVoice : Integer] : TSignal read GetMainQ        write SetMainQ;
  end;


  TSvfChamberlinChain = class
  private
    type TState = record
      FLevel      : TSignal;
      FFrequency  : TSignal;
      FQ          : TSignal;
    end;
    PState = ^TState;
  private
    FPolyphony  : Integer;
    FFilters    : array of TSvfChamberlin;
    FSampleRate : TSignal;
    FState      : array of TState;
  private
    function    GetCount: Integer;
    procedure   SetSampleRate( const aValue : TSignal);
    function    GetLevel     ( aVoice: Integer): TSignal;
    procedure   SetLevel     ( aVoice: Integer; const aValue: TSignal);
    function    GetFrequency ( aVoice: Integer): TSignal;
    procedure   SetFrequency ( aVoice: Integer; const aValue : TSignal);
    function    GetQ         ( aVoice: Integer): TSignal;
    procedure   SetQ         ( aVoice: Integer; const aValue : TSignal);
  public
    constructor Create( aPolyphony: Integer; aCount: Integer);
    destructor  Destroy;                                                                                       override;
    function    Process( aVoice: Integer; const anInput : TSignal): TSignal;                          virtual; abstract;
  public
    property    Polyphony                   : Integer read FPolyphony;
    property    Count                       : Integer read GetCount;
    property    SampleRate                  : TSignal read FSampleRate  write SetSampleRate;
    property    Level    [ aVoice: Integer] : TSignal read GetLevel     write SetLevel;
    property    Frequency[ aVoice: Integer] : TSignal read GetFrequency write SetFrequency;
    property    Q        [ aVoice: Integer] : TSignal read GetQ         write SetQ;
  end;


  TSvfChamberlinChainLowPass = class( TSvfChamberlinChain)
  public
    function    Process( aVoice: Integer; const anInput : TSignal): TSignal;                                   override;
  end;


  TSvfChamberlinChainBandPass = class( TSvfChamberlinChain)
  public
    function    Process( aVoice: Integer; const anInput : TSignal): TSignal;                                   override;
  end;


  TSvfChamberlinChainBandReject = class( TSvfChamberlinChain)
  public
    function    Process( aVoice: Integer; const anInput : TSignal): TSignal;                                   override;
  end;


  TSvfChamberlinChainHighPass = class( TSvfChamberlinChain)
  public
    function    Process( aVoice: Integer; const anInput : TSignal): TSignal;                                   override;
  end;


  TSvfChamberlinBank = class
  strict private
  const
    FILTER_COUNT = 17;
  private
    FPolyphony : Integer;
    FLocked    : Boolean;
    FFilters   : array of TSvfChamberlinChain;
    FFis       : Boolean;
  private
    function    GetCount: Integer;
    procedure   SetLevel    ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetFrequency( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetQ        ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetLevels   ( const aVoice: Integer; const aValue: TSignal);
    procedure   SetQs       ( const aVoice: Integer; const aValue: TSignal);
  public
    constructor Create( aPolyphony: Integer; aSampleRate: TSignal; aLength, aSteepness: Integer; aModeLow, aModeMid, aModeHigh: TSvfMode);
    destructor  Destroy;                                                                                       override;
    procedure   Process       ( aVoice: Integer; const anInput : TSignal; out aValue : TSignal);
    procedure   ProcessStereo ( aVoice: Integer; const anInput : TSignal; out aValueL, aValueR : TSignal);
    procedure   ProcessSplit  ( aVoice: Integer; const anInput : TSignal; var out1, out2, out3, out4, out5, out6, out7, out8, out9, out10, out11, out12, out13, out14, out15, out16, out17 : TSignal);
    procedure   ProcessCombine( aVoice: Integer; const anIn1, anIn2, anIn3, anIn4, anIn5, anIn6, anIn7, anIn8, anIn9, anIn10, anIn11, anIn12, anIn13, anIn14, anIn15, anIn16, anIn17: TSignal; var anOut: TSignal);
  public
    property    Polyphony                                  : Integer read FPolyphony;
    property    Count                                      : Integer read GetCount;
    property    Level    [ const aVoice, anIndex: Integer] : TSignal                 write SetLevel;
    property    Frequency[ const aVoice, anIndex: Integer] : TSignal                 write SetFrequency;
    property    Q        [ const aVoice, anIndex: Integer] : TSignal                 write SetQ;
    property    Levels   [ const aVoice         : Integer] : TSignal                 write SetLevels;
    property    Qs       [ const aVoice         : Integer] : TSignal                 write SetQs;
    property    FIs                                        : Boolean read FFis       write FFIs;
  end;


  TSvfFormantFilter = class
  private
    type TState = record
      FVowelIndex : TSignal;
    end;
    PState = ^ TState;
  private
    FPolyphony         : Integer;
    FFormantCollection : TFormantCollection;
    FLocked            : Boolean;
    FSampleRate        : TSignal;
    FFilters           : array of TSvfChamberlin;
    FLevels            : array of TSignal;
    FCurrentSet        : Integer;
    FFormantCount      : Integer;
    FVowel             : TVowel;
    FFis               : Boolean;
    FState             : array of TState;
  private
    function    GetSetCount : Integer;
    procedure   SetCurrentSet( anIndex: Integer);
    procedure   SetVowelIndex( aVoice: Integer; anIndex: TSignal);
    procedure   SetSampleRate( aValue: TSignal);
  private
    procedure   RecreateFilter;
    procedure   UpdateFilter( aVoice: Integer);
  public
    constructor Create( aPolyphony: Integer; aFormantCollection: TFormantCollection; aSampleRate: TSignal);
    destructor  Destroy;                                                                                       override;
    procedure   Process( aVoice: Integer; const anInput, aVowelIndex: TSignal; out aValue: TSignal);
  public
    property    Polyphony  : Integer read FPolyphony;
    property    SetCount   : Integer read GetSetCount;
    property    CurrentSet : Integer read FCurrentSet write SetCurrentSet;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    FIs        : Boolean read FFis        write FFIs;
  end;


  TSvfModalFilter = class
  private
    FPolyphony       : Integer;
    FModalCollection : TModalCollection;
    FMaxVoices       : Integer;
    FLocked          : Boolean;
    FSampleRate      : TSignal;
    FFilters         : array of TSvfChamberlinRelative;
    FLevels          : array of TSignal;
    FCurrent         : Integer;
    FFis             : Boolean;
  private
    function    GetCount       : Integer;
    function    GetFilterCount : Integer;
    function    GetName        : string;
    procedure   SetLevel     ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetFrequency ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetLFM       ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetQ         ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetMainQ     ( const aVoice: Integer;          const aValue: TSignal);
    procedure   SetLevels    ( const aVoice: Integer;          const aValue: TSignal);
    procedure   SetQs        ( const aVoice: Integer;          const aValue: TSignal);
    procedure   SetCurrent   ( anIndex: Integer);
    procedure   SetSampleRate( aValue : TSignal);
  private
    procedure   RecreateFilter;
  public
    constructor Create( aPolyphony: Integer; aModalCollection: TModalCollection; aSampleRate: TSignal; aMaxVoices : Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Process( aVoice: Integer; const anInput: TSignal; out aValue: TSignal);
  public
    property    Polyphony                                  : Integer read FPolyphony;
    property    Count                                      : Integer read GetCount;
    property    Current                                    : Integer read FCurrent       write SetCurrent;
    property    SampleRate                                 : TSignal read FSampleRate    write SetSampleRate;
    property    FilterCount                                : Integer read GetFilterCount;
    property    Name                                       : string  read GetName;
    property    Level    [ const aVoice, anIndex: Integer] : TSignal                     write SetLevel;
    property    Frequency[ const aVoice, anIndex: Integer] : TSignal                     write SetFrequency; // in Units
    property    LFM      [ const aVoice, anIndex: Integer] : TSignal                     write SetLFM;
    property    Q        [ const aVoice, anIndex: Integer] : TSignal                     write SetQ;
    property    Levels   [ const aVoice: Integer]          : TSignal                     write SetLevels;
    property    Qs       [ const aVoice: Integer]          : TSignal                     write SetQs;
    property    MainQ    [ const aVoice: Integer]          : TSignal                     write SetMainQ;
    property    FIs                                        : Boolean read FFis           write FFIs;
  end;


  TSvfChladniFilter = class
  private
    FPolyphony  : Integer;
    FMaxVoices  : Integer;
    FLocked     : Boolean;
    FSampleRate : TSignal;
    FFilters    : array of TSvfChamberlinRelative;
    FLevels     : array of TSignal;
    FFis        : Boolean;
  private
    function    GetFilterCount : Integer;
    procedure   SetLevel        ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetRelFrequency ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetQ            ( const aVoice, anIndex: Integer; const aValue: TSignal);
    procedure   SetFrequency    ( const aVoice: Integer; const aValue: TSignal);
    procedure   SetLFM          ( const aVoice: Integer; const aValue: TSignal);
    procedure   SetMainQ        ( const aVoice: Integer; const aValue: TSignal);
    procedure   SetLevels       ( const aVoice: Integer; const aValue: TSignal);
    procedure   SetQs           ( const aVoice: Integer; const aValue: TSignal);
    procedure   SetSampleRate   ( const aValue: TSignal);
  private
    procedure   RecreateFilter;
  public
    constructor Create( aPolyphony: Integer; aSampleRate: TSignal; aMaxVoices : Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Process( aVoice: Integer; const anInput: TSignal; out aValue: TSignal);
  public
    property    Polyphony                                     : Integer read FPolyphony;
    property    SampleRate                                    : TSignal read FSampleRate    write SetSampleRate;
    property    FilterCount                                   : Integer read GetFilterCount;
    property    Level       [ const aVoice, anIndex: Integer] : TSignal                     write SetLevel;
    property    RelFrequency[ const aVoice, anIndex: Integer] : TSignal                     write SetRelFrequency; // in Units
    property    Q           [ const aVoice, anIndex: Integer] : TSignal                     write SetQ;
    property    Frequency   [ const aVoice: Integer         ] : TSignal                     write SetFrequency;    // in Units
    property    LFM         [ const aVoice: Integer         ] : TSignal                     write SetLFM;
    property    Levels      [ const aVoice: Integer         ] : TSignal                     write SetLevels;
    property    Qs          [ const aVoice: Integer         ] : TSignal                     write SetQs;
    property    MainQ       [ const aVoice: Integer         ] : TSignal                     write SetMainQ;
    property    FIs                                           : Boolean read FFis           write FFIs;
  end;


implementation


const

  MIN_SAMPRATE = 44100.0;                 // Minimal sample rate, in Hz


{ ========
  TSvfChamberlin = class
  private
    type TState = record
      FInternalQ    : TSignal;
      FQ            : TSignal;
      FInternalFreq : TSignal;
      FFrequency    : TSignal;
      FDelay1       : TSignal;
      FDelay2       : TSignal;
    end;
    PState = ^ TState;
  private
    FPolyphony    : Integer;
    FSampleRate   : TSignal;
    FReciprocalSR : TSignal;
    FState        : array of TState;
  public
    property    Polyphony                    : Integer read FPolyphony;
    property    SampleRate                   : TSignal                 write SetSampleRate;
    property    Frequency [ aVoice: Integer] : TSignal                 write SetFrequency;
    property    Q         [ aVoice: Integer] : TSignal                 write SetQ;
  private
}

    procedure   TSvfChamberlin.FixFrequency( aVoice: Integer);
    var
      State  : PState;
      aValue : TSignal;
    begin
      State  := @ FState[ aVoice];
      aValue := State^.FFrequency;

      if   aValue >  FSampleRate / 6
      then aValue := FSampleRate / 6
      else if aValue < 0
      then aValue := 0;

      State^.FInternalFreq := 2.0 * LookupSine( 0.5 * aValue * FReciprocalSR);
    end;


    procedure   TSvfChamberlin.FixQ( aVoice: Integer);
    var
      State  : PState;
      aValue : TSignal;
    begin
      State  := @ FState[ aVoice];
      aValue := State^.FQ;

      if   aValue <  0.707
      then aValue := 0.707;

      State^.FInternalQ := 1.0 / aValue;
    end;


//  private

    procedure   TSvfChamberlin.SetSampleRate( const aValue : TSignal);
    var
      zz    : integer;
      aRate : TSignal;
    begin
      if   aValue <= MIN_SAMPRATE
      then aRate := MIN_SAMPRATE
      else aRate := aValue;

      if   aRate <> FSampleRate
      then begin
        FSampleRate   := aValue;
        FReciprocalSR := 1 / FSampleRate;

        for zz := 0 to Polyphony - 1
        do FixFrequency( zz);
      end;
    end;


    function    TSvfChamberlin.GetFrequency( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FFrequency;
    end;


    procedure   TSvfChamberlin.SetFrequency( aVoice: Integer; const aValue : TSignal);
    var
      State  : PState;
    begin
      State  := @ FState[ aVoice];

      if   aValue <> State^.FFrequency
      then begin
        State^.FFrequency := aValue;
        FixFrequency( aVoice);
      end;
    end;


    function    TSvfChamberlin.GetQ( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FQ;
    end;


    procedure   TSvfChamberlin.SetQ( aVoice: Integer; const aValue : TSignal);
    var
      State  : PState;
    begin
      State  := @ FState[ aVoice];

      if   aValue <> State^.FQ
      then begin
        State^.FQ := aValue;
        FixQ( aVoice);
      end;
    end;


//  public

    constructor TSvfChamberlin.Create( aPolyphony: Integer);
    var
      zz : Integer;
    begin
      inherited Create;
      FPolyphony := aPolyphony;
      SetLength( FState, Polyphony);
      SampleRate := MIN_SAMPRATE;               // First set sample rate, then Frequency

      for zz := 0 to Polyphony - 1
      do begin
        Frequency[ zz] := 329.6;                // in Hz
        Q        [ zz] :=   1.0;                // Q can be set at arbitrary time
        FixFrequency( zz);
        FixQ        ( zz);
      end;
    end;


    procedure   TSvfChamberlin.Process( aVoice: Integer; const anInput : TSignal; var aLowPass, aBandPass, aBandReject, aHighPass : TSignal);
    var
      State  : PState;
    begin
      State          := @ FState[ aVoice];
      aLowPass       := Normalize( State^.FDelay2 + State^.FInternalFreq * State^.FDelay1);
      aHighPass      := Normalize( anInput - aLowPass - State^.FInternalQ * State^.FDelay1);
      aBandPass      := Normalize( State^.FInternalFreq * aHighPass + State^.FDelay1);
      aBandReject    := Normalize( aHighPass + aLowPass);
      State^.FDelay1 := aBandPass;
      State^.FDelay2 := aLowPass;
    end;


{ ========
  TSvfChamberlinRelative = class
  private
    FInternalQ    : TSignal;
    FQ            : TSignal;
    FInternalFreq : TSignal;
    FFrequency    : TSignal;    // in Units
    FLFM          : TSignal;
    FSampleRate   : TSignal;
    FReciprocalSR : TSignal;
    FDelay1       : TSignal;
    FDelay2       : TSignal;
    FRelFrequency : TSignal;    // in Units
    FRelQ         : TSignal;
    FMainQ        : TSignal;
  public
    property    Frequency    : TSignal read FFrequency    write SetFrequency;
    property    LFM          : TSignal read FLFM          write SetLFM;
    property    SampleRate   : TSignal read FSampleRate   write SetSampleRate;
    property    Q            : TSignal read FQ            write SetQ;
    property    RelFrequency : TSignal read FRelFrequency write SetRelFrequency;
    property    RelQ         : TSignal read FRelQ         write SetRelQ;
    property    MainQ        : TSignal read FMainQ        write SetMainQ;
  private
}

    procedure   TSvfChamberlinRelative.FixFrequency( aVoice: Integer);
    var
      State  : PState;
      aValue : TSignal;
    begin
      State := @ FState[ aVoice];
      aValue := ( 1 + 0.5 * State^.FLFM) * LookupFrequency( State^.FFrequency + State^.FRelFrequency - NoteNumberToUnits( MiddleNote));

      if   aValue > FSampleRate / 6
      then aValue := FSampleRate / 6
      else if aValue < 0
      then aValue := 0;

      State^.FInternalFreq := 2.0 * LookupSine( 0.5 * aValue * FReciprocalSR);
    end;


    procedure   TSvfChamberlinRelative.FixQ( aVoice: Integer);
    var
      State  : PState;
      aValue : TSignal;
    begin
      State := @ FState[ aVoice];
      aValue := State^.FQ * State^.FRelQ * State^.FmainQ;

      if   aValue < 0.707
      then aValue := 0.707;

      State^.FInternalQ := 1.0 / aValue;
    end;


//  private

    procedure   TSvfChamberlinRelative.SetSampleRate( const aValue : TSignal);
    var
      zz    : Integer;
      aRate : TSignal;
    begin
      if   aValue <= MIN_SAMPRATE
      then aRate := MIN_SAMPRATE
      else aRate := aValue;

      if   aRate <> FSampleRate
      then begin
        FSampleRate   := aValue;
        FReciprocalSR := 1 / FSampleRate;

        for zz := 0 to Polyphony - 1
        do FixFrequency( zz);
      end;
    end;


    function    TSvfChamberlinRelative.GetFrequency( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FFrequency;
    end;


    procedure   TSvfChamberlinRelative.SetFrequency( aVoice: Integer; const aValue : TSignal);
    begin
      if   aValue <> FState[ aVoice].FFrequency
      then begin
        FState[ aVoice].FFrequency := aValue;
        FixFrequency( aVoice);
      end;
    end;


    function    TSvfChamberlinRelative.GetLFM( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FLFM;
    end;


    procedure   TSvfChamberlinRelative.SetLFM( aVoice: Integer; const aValue : TSignal);
    begin
      if   aValue <> FState[ aVoice].FLFM
      then begin
        FState[ aVoice].FLFM := aValue;
        FixFrequency( aVoice);
      end;
    end;


    function    TSvfChamberlinRelative.GetQ( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FQ;
    end;


    procedure   TSvfChamberlinRelative.SetQ( aVoice: Integer; const aValue : TSignal);
    begin
      if   aValue <> FState[ aVoice].FQ
      then begin
        FState[ aVoice].FQ := aValue;
        FixQ( aVoice);
      end;
    end;


    function    TSvfChamberlinRelative.GetRelFrequency( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FRelFrequency;
    end;


    procedure   TSvfChamberlinRelative.SetRelFrequency( aVoice: Integer; const aValue : TSignal);
    begin
      if   aValue <> FState[ aVoice].FRelFrequency
      then begin
        FState[ aVoice].FRelFrequency := aValue;
        FixFrequency( aVoice);
      end;
    end;


    function    TSvfChamberlinRelative.GetRelQ( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FRelQ;
    end;


    procedure   TSvfChamberlinRelative.SetRelQ( aVoice: Integer; const aValue : TSignal);
    begin
      if   aValue <> FState[ aVoice].FRelQ
      then begin
        FState[ aVoice].FRelQ := aValue;
        FixQ( aVoice);
      end;
    end;


    function    TSvfChamberlinRelative.GetMainQ( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FMainQ;
    end;


    procedure   TSvfChamberlinRelative.SetMainQ( aVoice: Integer; const aValue : TSignal);
    begin
      if   aValue <> FState[ aVoice].FMainQ
      then begin
        FState[ aVoice].FMainQ := aValue;
        FixQ( aVoice);
      end;
    end;


//  public

    constructor TSvfChamberlinRelative.Create( aPolyphony: Integer);
    var
      zz : Integer;
    begin
      inherited Create;
      FPolyphony := aPolyphony;
      SetLength( FState, Polyphony);

      for zz := 0 to Polyphony - 1
      do begin
        RelFrequency[ zz] := 1.0;
        RelQ        [ zz] := 1.0;
        MainQ       [ zz] := 1.0;
      end;

      SampleRate    := MIN_SAMPRATE;         // First set sample rate, then Frequency

      for zz := 0 to Polyphony - 1
      do begin
        Frequency[ zz] := 329.6;                // in Hz
        Q        [ zz] :=   1.0;                // Q can be set at arbitrary time
        FixFrequency( zz);
        FixQ        ( zz);
      end;
    end;


    procedure   TSvfChamberlinRelative.Process( aVoice: Integer; const anInput : TSignal; var aLowPass, aBandPass, aBandReject, aHighPass : TSignal);
    var
      State : PState;
    begin
      State := @ FState[ aVoice];
      aLowPass       := Normalize( State^.FDelay2 + State^.FInternalFreq * State^.FDelay1);
      aHighPass      := Normalize( anInput - aLowPass - State^.FInternalQ * State^.FDelay1);
      aBandPass      := Normalize( State^.FInternalFreq * aHighPass + State^.FDelay1);
      aBandReject    := Normalize( aHighPass + aLowPass);
      State^.FDelay1 := aBandPass;
      State^.FDelay2 := aLowPass;
    end;


{ ========
  TSvfChamberlinChain = class
  private
    type TState = record
      FLevel      : TSignal;
      FFrequency  : TSignal;
      FQ          : TSignal;
    end;
    PState = ^TState;
  private
    FPolyphony  : Integer;
    FFilters    : array of TSvfChamberlin;
    FSampleRate : TSignal;
    FState      : array of TState;
  public
    property    Polyphony                   : Integer read FPolyphony;
    property    Count                       : Integer read GetCount;
    property    SampleRate                  : TSignal read FSampleRate  write SetSampleRate;
    property    Level    [ aVoice: Integer] : TSignal read GetLevel     write SetLevel;
    property    Frequency[ aVoice: Integer] : TSignal read GetFrequency write SetFrequency;
    property    Q        [ aVoice: Integer] : TSignal read GetQ         write SetQ;
  private
}

    function    TSvfChamberlinChain.GetCount: Integer;
    begin
      Result := Length( FFilters);
    end;


    procedure   TSvfChamberlinChain.SetSampleRate( const aValue : TSignal);
    var
      i : Integer;
    begin
      if   aValue <> FSampleRate
      then begin
        FSampleRate := aValue;

        for i:= 0 to Count - 1
        do FFilters[ i].SampleRate := FSampleRate;
      end;
    end;


    function    TSvfChamberlinChain.GetLevel( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FLevel;
    end;


    procedure   TSvfChamberlinChain.SetLevel( aVoice: Integer; const aValue: TSignal);
    begin
      FState[ aVoice].FLevel := aValue;
    end;


    function    TSvfChamberlinChain.GetFrequency( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FFrequency;
    end;


    procedure   TSvfChamberlinChain.SetFrequency( aVoice: Integer; const aValue : TSignal);
    var
      i : Integer;
    begin
      if   aValue <> Frequency[ aVoice]
      then begin
        FState[ aVoice].FFrequency := aValue;

        for i:= 0 to Count - 1
        do FFilters[ i].Frequency[ aVoice] := FState[ aVoice].FFrequency;
      end;
    end;


    function    TSvfChamberlinChain.GetQ( aVoice: Integer): TSignal;
    begin
      Result := FState[ aVoice].FQ;
    end;


    procedure   TSvfChamberlinChain.SetQ( aVoice: Integer; const aValue : TSignal);
    var
      i : Integer;
    begin
      if   aValue <> FState[ aVoice].FQ
      then begin
        FState[ aVoice].FQ := aValue;

        for i:= 0 to Count - 1
        do FFilters[ i].Q[ aVoice] := FState[ aVoice].FQ / Count;
      end;
    end;


//  public

    constructor TSvfChamberlinChain.Create( aPolyphony: Integer; aCount: Integer);
    var
      i : Integer;
    begin
      inherited Create;
      FPolyphony := aPolyphony;
      SetLength( FState, Polyphony);
      SetLength( FFilters, aCount);

      for i := 0 to Count - 1
      do FFilters[ i] := TSvfChamberlin.Create( Polyphony);
    end;


    destructor  TSvfChamberlinChain.Destroy; // override;
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do FreeAndNil( FFilters[ i]);

      inherited;
    end;


{ ========
  TSvfChamberlinChainLowPass = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainLowPass.Process( aVoice: Integer; const anInput : TSignal): TSignal; // override;
    var
      aBandPass   : TSignal;
      aBandReject : TSignal;
      aHighPass   : TSignal;
      i           : Integer;
    begin
      FFilters[ 0].Process( aVoice, Normalize( FState[ aVoice].FLevel * anInput), Result, aBandPass, aBandReject, aHighPass);

      for i := 1 to Count - 1
      do FFilters[ i].Process( aVoice, Result, Result, aBandPass, aBandReject, aHighPass);
    end;


{ ========
  TSvfChamberlinChainBandPass = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainBandPass.Process( aVoice: Integer; const anInput : TSignal): TSignal; // override;
    var
      aLowPass    : TSignal;
      aBandReject : TSignal;
      aHighPass   : TSignal;
      i           : Integer;
    begin
      FFilters[ 0].Process( aVoice, Normalize( FState[ aVoice].FLevel * anInput), aLowPass, Result, aBandReject, aHighPass);

      for i := 1 to Count - 1
      do FFilters[ i].Process( aVoice, Result, aLowPass, Result, aBandReject, aHighPass);
    end;


{ ========
  TSvfChamberlinChainBandReject = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainBandReject.Process( aVoice: Integer; const anInput : TSignal): TSignal; // override;
    var
      aLowPass  : TSignal;
      aBandPass : TSignal;
      aHighPass : TSignal;
      i         : Integer;
    begin
      FFilters[ 0].Process( aVoice, Normalize( FState[ aVoice].FLevel * anInput), aLowPass, aBandPass, Result, aHighPass);

      for i := 1 to Count - 1
      do FFilters[ i].Process( aVoice, Result, aLowPass, aBandPass, Result, aHighPass);
    end;


{ ========
  TSvfChamberlinChainHighPass = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainHighPass.Process( aVoice: Integer; const anInput : TSignal): TSignal; // override;
    var
      aLowPass    : TSignal;
      aBandPass   : TSignal;
      aBandReject : TSignal;
      i           : Integer;
    begin
      FFilters[ 0].Process( aVoice, Normalize( FState[ aVoice].FLevel * anInput), aLowPass, aBandPass, aBandReject, Result);

      for i := 1 to Count - 1
      do FFilters[ i].Process( aVoice, Result, aLowPass, aBandPass, aBandReject, Result);
    end;


{ ========
  TSvfChamberlinBank = class
  private
    FPolyphony : Integer;
    FLocked    : Boolean;
    FFilters   : array of TSvfChamberlinChain;
    FFis       : Boolean;
  public
    property    Polyphony                                  : Integer read FPolyphony;
    property    Count                                      : Integer read GetCount;
    property    Level    [ const aVoice, anIndex: Integer] : TSignal                 write SetLevel;
    property    Frequency[ const aVoice, anIndex: Integer] : TSignal                 write SetFrequency;
    property    Q        [ const aVoice, anIndex: Integer] : TSignal                 write SetQ;
    property    Levels   [ const aVoice         : Integer] : TSignal                 write SetLevels;
    property    Qs       [ const aVoice         : Integer] : TSignal                 write SetQs;
    property    FIs                                        : Boolean read FFis       write FFIs;
  private
}

    function    TSvfChamberlinBank.GetCount: Integer;
    begin
      Result := Length( FFilters);
    end;


    procedure   TSvfChamberlinBank.SetLevel( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      FFilters[ anIndex].Level[ aVoice] := aValue;
    end;


    procedure   TSvfChamberlinBank.SetFrequency( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      FFilters[ anIndex].Frequency[ aVoice] := aValue;
    end;


    procedure   TSvfChamberlinBank.SetQ( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      FFilters[ anIndex].Q[ aVoice] := aValue;
    end;


    procedure   TSvfChamberlinBank.SetLevels( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do FFilters[ i].Level[ aVoice] := aValue;
    end;


    procedure   TSvfChamberlinBank.SetQs( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do FFilters[ i].Q[ aVoice] := aValue;
    end;


//  public

    constructor TSvfChamberlinBank.Create( aPolyphony: Integer; aSampleRate: TSignal; aLength, aSteepness: Integer; aModeLow, aModeMid, aModeHigh: TSvfMode);

      function MakeFilter( aMode: TSvfMode; aSteepness: Integer): TSvfChamberlinChain;
      begin
        case aMode of
          svfmLowPass  : Result := TSvfChamberlinChainLowPass   .Create( Polyphony, aSteepness);
          svfmBandPass : Result := TSvfChamberlinChainBandPass  .Create( Polyphony, aSteepness);
          svfmHighPass : Result := TSvfChamberlinChainHighPass  .Create( Polyphony, aSteepness);
          else           Result := TSvfChamberlinChainBandReject.Create( Polyphony, aSteepness);
        end;
      end;

    var
      zz : Integer;
      i  : Integer;
    begin
      Assert( aSteepness > 0);
      FLocked := true;
      inherited Create;
      FPolyphony := aPolyphony;
      SetLength( FFilters, aLength);;

      for i := 0 to aLength - 1
      do begin
        if   i = 0
        then FFilters[ i] := MakeFilter( aModeLow , aSteepness)
        else if i = aLength - 1
        then FFilters[ i] := MakeFilter( aModeHigh, aSteepness)
        else FFilters[ i] := MakeFilter( aModeMid , aSteepness);

        for zz := 0 to Polyphony - 1
        do begin
          Level    [ zz, i] := 0;
          Frequency[ zz, i] := 329.6;
        end;

        FFilters[ i].SampleRate := aSampleRate;
      end;

      for zz := 0 to Polyphony - 1
      do Qs[ zz] := 0.5;

      FLocked := False;
    end;


    destructor  TSvfChamberlinBank.Destroy; // override;
    var
      i: Integer;
    begin
      FLocked := True;

      for i := 0 to Count - 1
      do FreeAndNil( FFilters[ i]);

      inherited;
    end;


    procedure   TSvfChamberlinBank.Process( aVoice: Integer; const anInput : TSignal; out aValue : TSignal);
    var
      i : Integer;
    begin
      aValue := 0.0;

      if   not FLocked
      then begin
        for i := 0 to Count - 1
        do begin
          if   FIs
          then begin
            if   Odd( i)
            then aValue := Normalize( aValue - FFilters[ i].Process( aVoice, anInput))
            else aValue := Normalize( aValue + FFilters[ i].Process( aVoice, anInput));
          end
          else aValue := Normalize( aValue + FFilters[ i].Process( aVoice, anInput));
        end;
      end;
    end;


    procedure   TSvfChamberlinBank.ProcessStereo( aVoice: Integer; const anInput : TSignal; out aValueL, aValueR : TSignal);
    var
      i : Integer;
    begin
      aValueL := 0.0;
      aValueR := 0.0;

      if   not FLocked
      then begin
        for i := 0 to Count - 1
        do begin
          if   Odd( i)
          then aValueL := Normalize( aValueL + FFilters[ i].Process( aVoice, anInput))
          else aValueR := Normalize( aValueR + FFilters[ i].Process( aVoice, anInput));
        end;
      end;
    end;


    procedure   TSvfChamberlinBank.ProcessSplit( aVoice: Integer; const anInput : TSignal; var out1, out2, out3, out4, out5, out6, out7, out8, out9, out10, out11, out12, out13, out14, out15, out16, out17 : TSignal);
    begin
      if   FLocked
      then begin
        out1  := 0.0;
        out2  := 0.0;
        out3  := 0.0;
        out4  := 0.0;
        out5  := 0.0;
        out7  := 0.0;
        out7  := 0.0;
        out8  := 0.0;
        out9  := 0.0;
        out10 := 0.0;
        out11 := 0.0;
        out12 := 0.0;
        out13 := 0.0;
        out14 := 0.0;
        out15 := 0.0;
        out16 := 0.0;
        out17 := 0.0;
      end
      else begin
        if   FIs
        then begin
          out1  :=   Normalize( FFilters[  0].Process( aVoice, anInput));
          out2  := - Normalize( FFilters[  1].Process( aVoice, anInput));
          out3  :=   Normalize( FFilters[  2].Process( aVoice, anInput));
          out4  := - Normalize( FFilters[  3].Process( aVoice, anInput));
          out5  :=   Normalize( FFilters[  4].Process( aVoice, anInput));
          out7  := - Normalize( FFilters[  5].Process( aVoice, anInput));
          out7  :=   Normalize( FFilters[  6].Process( aVoice, anInput));
          out8  := - Normalize( FFilters[  7].Process( aVoice, anInput));
          out9  :=   Normalize( FFilters[  8].Process( aVoice, anInput));
          out10 := - Normalize( FFilters[  9].Process( aVoice, anInput));
          out11 :=   Normalize( FFilters[ 10].Process( aVoice, anInput));
          out12 := - Normalize( FFilters[ 11].Process( aVoice, anInput));
          out13 :=   Normalize( FFilters[ 12].Process( aVoice, anInput));
          out14 := - Normalize( FFilters[ 13].Process( aVoice, anInput));
          out15 :=   Normalize( FFilters[ 14].Process( aVoice, anInput));
          out16 := - Normalize( FFilters[ 15].Process( aVoice, anInput));
          out17 :=   Normalize( FFilters[ 16].Process( aVoice, anInput));
        end
        else begin
          out1  :=   Normalize( FFilters[  0].Process( aVoice, anInput));
          out2  :=   Normalize( FFilters[  1].Process( aVoice, anInput));
          out3  :=   Normalize( FFilters[  2].Process( aVoice, anInput));
          out4  :=   Normalize( FFilters[  3].Process( aVoice, anInput));
          out5  :=   Normalize( FFilters[  4].Process( aVoice, anInput));
          out7  :=   Normalize( FFilters[  5].Process( aVoice, anInput));
          out7  :=   Normalize( FFilters[  6].Process( aVoice, anInput));
          out8  :=   Normalize( FFilters[  7].Process( aVoice, anInput));
          out9  :=   Normalize( FFilters[  8].Process( aVoice, anInput));
          out10 :=   Normalize( FFilters[  9].Process( aVoice, anInput));
          out11 :=   Normalize( FFilters[ 10].Process( aVoice, anInput));
          out12 :=   Normalize( FFilters[ 11].Process( aVoice, anInput));
          out13 :=   Normalize( FFilters[ 12].Process( aVoice, anInput));
          out14 :=   Normalize( FFilters[ 13].Process( aVoice, anInput));
          out15 :=   Normalize( FFilters[ 14].Process( aVoice, anInput));
          out16 :=   Normalize( FFilters[ 15].Process( aVoice, anInput));
          out17 :=   Normalize( FFilters[ 16].Process( aVoice, anInput));
        end;
      end;
    end;


    procedure   TSvfChamberlinBank.ProcessCombine( aVoice: Integer; const anIn1, anIn2, anIn3, anIn4, anIn5, anIn6, anIn7, anIn8, anIn9, anIn10, anIn11, anIn12, anIn13, anIn14, anIn15, anIn16, anIn17: TSignal; var anOut: TSignal);
    begin
      if   FLocked
      then anOut  := 0.0
      else begin
        if FIs
        then
          anOut :=
            + Normalize( FFilters[  0].Process( aVoice, anIn1 ))
            - Normalize( FFilters[  1].Process( aVoice, anIn2 ))
            + Normalize( FFilters[  2].Process( aVoice, anIn3 ))
            - Normalize( FFilters[  3].Process( aVoice, anIn4 ))
            + Normalize( FFilters[  4].Process( aVoice, anIn5 ))
            - Normalize( FFilters[  5].Process( aVoice, anIn6 ))
            + Normalize( FFilters[  6].Process( aVoice, anIn7 ))
            - Normalize( FFilters[  7].Process( aVoice, anIn8 ))
            + Normalize( FFilters[  8].Process( aVoice, anIn9 ))
            - Normalize( FFilters[  9].Process( aVoice, anIn10))
            + Normalize( FFilters[ 10].Process( aVoice, anIn11))
            - Normalize( FFilters[ 11].Process( aVoice, anIn12))
            + Normalize( FFilters[ 12].Process( aVoice, anIn13))
            - Normalize( FFilters[ 13].Process( aVoice, anIn14))
            + Normalize( FFilters[ 14].Process( aVoice, anIn15))
            - Normalize( FFilters[ 15].Process( aVoice, anIn16))
            + Normalize( FFilters[ 16].Process( aVoice, anIn17))
        else
          anOut :=
            + Normalize( FFilters[  0].Process( aVoice, anIn1 ))
            + Normalize( FFilters[  1].Process( aVoice, anIn2 ))
            + Normalize( FFilters[  2].Process( aVoice, anIn3 ))
            + Normalize( FFilters[  3].Process( aVoice, anIn4 ))
            + Normalize( FFilters[  4].Process( aVoice, anIn5 ))
            + Normalize( FFilters[  5].Process( aVoice, anIn6 ))
            + Normalize( FFilters[  6].Process( aVoice, anIn7 ))
            + Normalize( FFilters[  7].Process( aVoice, anIn8 ))
            + Normalize( FFilters[  8].Process( aVoice, anIn9 ))
            + Normalize( FFilters[  9].Process( aVoice, anIn10))
            + Normalize( FFilters[ 10].Process( aVoice, anIn11))
            + Normalize( FFilters[ 11].Process( aVoice, anIn12))
            + Normalize( FFilters[ 12].Process( aVoice, anIn13))
            + Normalize( FFilters[ 13].Process( aVoice, anIn14))
            + Normalize( FFilters[ 14].Process( aVoice, anIn15))
            + Normalize( FFilters[ 15].Process( aVoice, anIn16))
            + Normalize( FFilters[ 16].Process( aVoice, anIn17));
        end;
    end;


{ ========
  TSvfFormantFilter = class
  private
    type TState = record
      FVowelIndex : TSignal;
    end;
    PState = ^ TState;
  private
    FPolyphony         : Integer;
    FFormantCollection : TFormantCollection;
    FLocked            : Boolean;
    FSampleRate        : TSignal;
    FFilters           : array of TSvfChamberlin;
    FLevels            : array of TSignal;
    FCurrentSet        : Integer;
    FFormantCount      : Integer;
    FVowel             : TVowel;
    FFis               : Boolean;
    FState             : array of TState;
  public
    property    SetCount   : Integer read GetSetCount;
    property    CurrentSet : Integer read FCurrentSet write SetCurrentSet;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    FIs        : Boolean read FFis        write FFIs;
  private
}

    function    TSvfFormantFilter.GetSetCount: Integer;
    begin
      if   Assigned( FFormantCollection)
      then Result := FFormantCollection.SetCount
      else Result := 0;
    end;


    procedure   TSvfFormantFilter.SetCurrentSet( anIndex: Integer);
    begin
      if   SetCount > 0
      then begin
        anIndex := Clip( anIndex, 0, SetCount - 1);

        if   anIndex <> FCurrentSet
        then begin
          FCurrentSet := anIndex;
          RecreateFilter;
        end;
      end;
    end;


    procedure   TSvfFormantFilter.SetVowelIndex( aVoice: Integer; anIndex: TSignal);
    begin
      if   Assigned( FFormantCollection) and ( SetCount > 0)
      then begin
        if   anIndex <> FState[ aVoice].FVowelIndex
        then begin
          FState[ aVoice].FVowelIndex := anIndex;
          FFormantCollection.Interpolate( FCurrentSet, FState[ aVoice].FVowelIndex, FVowel);
          UpdateFilter( aVoice);
        end;
      end;
    end;


    procedure   TSvfFormantFilter.SetSampleRate( aValue: TSignal);
    var
      i : Integer;
    begin
      if   aValue <> FSampleRate
      then begin
        FSampleRate := aValue;

        for i := 0 to Length( FFilters) - 1
        do FFilters[ i].SampleRate := FSampleRate;
      end;
    end;


//  private

    procedure   TSvfFormantFilter.RecreateFilter;
    var
      zz : Integer;
      i  : Integer;
    begin
      FLocked := True;

      try
        for i := 0 to Length( FFilters) - 1
        do FreeAndNil( FFilters[ i]);

        if   SetCount > 0
        then FFormantCount := FFormantCollection.FormantCount[ FCurrentSet, 0]
        else FFormantCount := 0;

        SetLength( FFilters       , FFormantCount);
        SetLength( FLevels        , FFormantCount);
        SetLength( FVowel.Formants, FFormantCount);

        for i := 0 to Length( FFilters) - 1
        do FFilters[ i] := TSvfChamberlin.Create( Polyphony);

        for zz := 0 to Polyphony - 1
        do begin
          FState[ zz] . FVowelIndex := -1;
          SetVowelIndex( zz, 0);
        end;
      finally
        FLocked := False;
      end;
    end;


    procedure   TSvfFormantFilter.UpdateFilter( aVoice: Integer);
    var
      i : Integer;
    begin
      for i := 0 to FFormantCount - 1
      do begin
        FFilters[ i].Frequency[ aVoice] := FVowel.Formants[ i].Freg;
        FFilters[ i].Q        [ aVoice] := FVowel.Formants[ i].Q;
        FLevels [ i]                    := FVowel.Formants[ i].LinLevel;
      end;
    end;


//  public

    constructor TSvfFormantFilter.Create( aPolyphony: Integer; aFormantCollection: TFormantCollection; aSampleRate: TSignal);
    var
      zz : Integer;
    begin
      inherited Create;
      FLocked            := True;
      FPolyphony         := aPolyphony;
      SetLength( FState, Polyphony);
      FVowel.Name1       := '';
      FVowel.Name2       := '';
      FFormantCount      :=  0;
      FCurrentSet        := -1;

      for zz := 0 to Polyphony - 1
      do FState[ zz].FVowelIndex := -1;

      FFormantCollection := aFormantCollection;
      CurrentSet         := 0;                     // Will clear FLocked and will (re)create the filter
      SampleRate         := aSampleRate;

      for zz := 0 to Polyphony - 1
      do SetVowelIndex( zz, 0);                    // Will update the filter
    end;


    destructor  TSvfFormantFilter.Destroy; // override;
    var
      i : Integer;
    begin
      FLocked := True;

      for i := 0 to Length( FFilters) - 1
      do FreeAndNil( FFilters[ i]);

      inherited;
    end;


    procedure   TSvfFormantFilter.Process( aVoice: Integer; const anInput, aVowelIndex: TSignal; out aValue: TSignal);
    var
      i           : Integer;
      aLowPass    : TSignal; // note : need to use distinct variables for all params
      aBandPass   : TSignal; //        passed to FFilters[ i].Process, or odd stuff
      aBandReject : TSignal; //        will happen (even when the values are not being
      aHighPass   : TSignal; //        used).
    begin
      aValue := 0.0;

      if   not FLocked
      then begin
        SetVowelIndex( aVoice, aVowelIndex);

        for i := 0 to Length( FFilters) - 1
        do begin
          FFilters[ i].Process( aVoice, anInput, aLowPass, aBandPass, aBandReject, aHighPass);
          aValue := Normalize( aValue + FLevels[ i] * aBandPass);
        end;
      end;
    end;


{ ========
  TSvfModalFilter = class
  private
    FModalCollection : TModalCollection;
    FMaxVoices       : Integer;
    FLocked          : Boolean;
    FSampleRate      : TSignal;
    FFilters         : array of TSvfChamberlinRelative;
    FLevels          : array of TSignal;
    FCurrent         : Integer;
    FFis             : Boolean;
  public
    property    Polyphony                                  : Integer read FPolyphony;
    property    Count                                      : Integer read GetCount;
    property    Current                                    : Integer read FCurrent       write SetCurrent;
    property    SampleRate                                 : TSignal read FSampleRate    write SetSampleRate;
    property    FilterCount                                : Integer read GetFilterCount;
    property    Name                                       : string  read GetName;
    property    Level    [ const aVoice, anIndex: Integer] : TSignal                     write SetLevel;
    property    Frequency[ const aVoice, anIndex: Integer] : TSignal                     write SetFrequency; // in Units
    property    LFM      [ const aVoice, anIndex: Integer] : TSignal                     write SetLFM;
    property    Q        [ const aVoice, anIndex: Integer] : TSignal                     write SetQ;
    property    Levels   [ const aVoice: Integer]          : TSignal                     write SetLevels;
    property    Qs       [ const aVoice: Integer]          : TSignal                     write SetQs;
    property    MainQ    [ const aVoice: Integer]          : TSignal                     write SetMainQ;
    property    FIs                                        : Boolean read FFis           write FFIs;
  private
}

    function    TSvfModalFilter.GetCount: Integer;
    begin
      if   Assigned( FModalCollection)
      then Result := FModalCollection.SetCount
      else Result := 0;
    end;


    function    TSvfModalFilter.GetFilterCount: Integer;
    begin
      Result := Length( FFilters);
    end;


    function    TSvfModalFilter.GetName: string;
    begin
      if   Assigned( FModalCollection)
      then Result := FModalCollection.Name[ FCurrent]
      else Result := '<no modals>';
    end;


    procedure   TSvfModalFilter.SetLevel( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      // Abs
      FLevels[ anIndex] := aValue;
    end;


    procedure   TSvfModalFilter.SetFrequency( const aVoice, anIndex: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Rel, exp
      for i := 0 to FilterCount - 1
      do FFilters[ i].RelFrequency[ aVoice] := aValue;
    end;


    procedure   TSvfModalFilter.SetLFM( const aVoice, anIndex: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      for i := 0 to FilterCount - 1
      do FFilters[ i].LFM[ aVoice] := aValue;
    end;


    procedure   TSvfModalFilter.SetQ( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      // Rel, lin
      FFilters[ anIndex].RelQ[ aVoice] := aValue;
    end;


    procedure   TSvfModalFilter.SetMainQ( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Rel, lin
      for i := 0 to FilterCount - 1
      do FFilters[ i].MainQ[ aVoice] := aValue;
    end;


    procedure   TSvfModalFilter.SetLevels( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Abs
      for i := 0 to FilterCount - 1
      do FLevels[ i] := aValue;
    end;


    procedure   TSvfModalFilter.SetQs( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Rel
      for i := 0 to FilterCount - 1
      do FFilters[ i].RelQ[ aVoice] := aValue;
    end;


    procedure   TSvfModalFilter.SetCurrent( anIndex: Integer);
    begin
      if   Count > 0
      then begin
        anIndex := Clip( anIndex, 0, Count - 1);

        if   anIndex <> FCurrent
        then begin
          FCurrent := anIndex;
          RecreateFilter;
        end;
      end;
    end;


    procedure   TSvfModalFilter.SetSampleRate( aValue: TSignal);
    var
      i : Integer;
    begin
      if   aValue <> FSampleRate
      then begin
        FSampleRate := aValue;

        for i := 0 to FilterCount - 1
        do FFilters[ i].SampleRate := FSampleRate;
      end;
    end;


//  private

    procedure   TSvfModalFilter.RecreateFilter;
    var
      zz     : Integer;
      i      : Integer;
      aCount : Integer;
    begin
      FLocked := True;

      try
        for i := 0 to FilterCount - 1
        do FreeAndNil( FFilters[ i]);

        if   Count > 0
        then aCount := Min( FModalCollection.FormantCount[ FCurrent], FMaxVoices)
        else aCount := 0;

        SetLength( FFilters, aCount);
        SetLength( FLevels , aCount);

        for i := 0 to FilterCount - 1
        do begin
          FFilters[ i]            := TSvfChamberlinRelative.Create( Polyphony);
          FFilters[ i].SampleRate := SampleRate;

          for zz := 0 to Polyphony - 1
          do begin
            FFilters[ i].Frequency[ zz]  := FrequencyToUnits( FModalCollection.Frequency[ FCurrent, i]);
            FFilters[ i].Q        [ zz]  := FModalCollection.Q[ FCurrent, i];
          end;

          FLevels[ i] := 1.0;
        end;
      finally
        FLocked := False;
      end;
    end;


//  public

    constructor TSvfModalFilter.Create( aPolyphony: Integer; aModalCollection: TModalCollection; aSampleRate: TSignal; aMaxVoices : Integer);
    begin
      inherited Create;
      FLocked          := True;
      FPolyphony       := aPolyphony;
      FCurrent         := -1;
      FModalCollection := aModalCollection;
      FMaxVoices       := aMaxVoices;
      SampleRate       := aSampleRate;
      Current          := 0; // Will clear FLocked and will (re)create the filter, needs sample rate to be set before
    end;


    destructor  TSvfModalFilter.Destroy; // override;
    var
      i : Integer;
    begin
      FLocked := True;

      for i := 0 to Length( FFilters) - 1
      do FreeAndNil( FFilters[ i]);

      inherited;
    end;


    procedure   TSvfModalFilter.Process( aVoice: Integer; const anInput: TSignal; out aValue: TSignal);
    var
      i           : Integer;
      aLowPass    : TSignal; // note : need to use distinct variables for all params
      aBandPass   : TSignal; //        passed to FFilters[ i].Process, or odd stuff
      aBandReject : TSignal; //        will happen (even when the values are not being
      aHighPass   : TSignal; //        used).
    begin
      aValue := 0.0;

      if   not FLocked
      then begin
        for i := 0 to Length( FFilters) - 1
        do begin
          FFilters[ i].Process( aVoice, anInput, aLowPass, aBandPass, aBandReject, aHighPass);

          if   FIs
          then begin
            if   Odd( i)
            then aValue := Normalize( aValue - FLevels[ i] * aBandPass)
            else aValue := Normalize( aValue + FLevels[ i] * aBandPass);
          end
          else aValue := Normalize( aValue + FLevels[ i] * aBandPass);
        end;
      end;
    end;


{ ========
  TSvfChladniFilter = class
  private
    FPolyphony  : Integer;
    FMaxVoices  : Integer;
    FLocked     : Boolean;
    FSampleRate : TSignal;
    FFilters    : array of TSvfChamberlinRelative;
    FLevels     : array of TSignal;
    FFis        : Boolean;
  public
    property    Polyphony                                     : Integer read FPolyphony;
    property    SampleRate                                    : TSignal read FSampleRate    write SetSampleRate;
    property    FilterCount                                   : Integer read GetFilterCount;
    property    Level       [ const aVoice, anIndex: Integer] : TSignal                     write SetLevel;
    property    RelFrequency[ const aVoice, anIndex: Integer] : TSignal                     write SetRelFrequency; // in Units
    property    Q           [ const aVoice, anIndex: Integer] : TSignal                     write SetQ;
    property    Frequency   [ const aVoice: Integer         ] : TSignal                     write SetFrequency;    // in Units
    property    LFM         [ const aVoice: Integer         ] : TSignal                     write SetLFM;
    property    Levels      [ const aVoice: Integer         ] : TSignal                     write SetLevels;
    property    Qs          [ const aVoice: Integer         ] : TSignal                     write SetQs;
    property    MainQ       [ const aVoice: Integer         ] : TSignal                     write SetMainQ;
    property    FIs                                           : Boolean read FFis           write FFIs;
  private
}

    function    TSvfChladniFilter.GetFilterCount: Integer;
    begin
      Result := Length( FFilters);
    end;


    procedure   TSvfChladniFilter.SetLevel( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      // Abs
      FLevels[ aVoice] := aValue;
    end;


    procedure   TSvfChladniFilter.SetRelFrequency( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      // Rel, exp
      FFilters[ anIndex].RelFrequency[ aVoice] := aValue;
    end;


    procedure   TSvfChladniFilter.SetQ( const aVoice, anIndex: Integer; const aValue: TSignal);
    begin
      // Rel, lin
      FFilters[ anIndex].RelQ[ aVoice] := aValue;
    end;


    procedure   TSvfChladniFilter.SetFrequency( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Abs, exp
      for i := 0 to FilterCount - 1
      do FFilters[ i].Frequency[ aVoice] := aValue;
    end;


    procedure   TSvfChladniFilter.SetLFM( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      for i := 0 to FilterCount - 1
      do FFilters[ i].LFM[ aVoice] := aValue;
    end;


    procedure   TSvfChladniFilter.SetMainQ( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Rel, lin
      for i := 0 to FilterCount - 1
      do FFilters[ i].MainQ[ aVoice] := aValue;
    end;


    procedure   TSvfChladniFilter.SetLevels( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Abs
      for i := 0 to FilterCount - 1
      do FLevels[ i] := aValue;
    end;


    procedure   TSvfChladniFilter.SetQs( const aVoice: Integer; const aValue: TSignal);
    var
      i : Integer;
    begin
      // Rel
      for i := 0 to FilterCount - 1
      do FFilters[ i].RelQ[ aVoice] := aValue;
    end;


    procedure   TSvfChladniFilter.SetSampleRate( const aValue: TSignal);
    var
      i : Integer;
    begin
      if   aValue <> FSampleRate
      then begin
        FSampleRate := aValue;

        for i := 0 to FilterCount - 1
        do FFilters[ i].SampleRate := FSampleRate;
      end;
    end;


//  private

    procedure   TSvfChladniFilter.RecreateFilter;
    var
      zz     : Integer;
      i      : Integer;
      aCount : Integer;
    begin
      FLocked := True;

      try
        for i := 0 to FilterCount - 1
        do FreeAndNil( FFilters[ i]);

        aCount := FMaxVoices;

        SetLength( FFilters, aCount);
        SetLength( FLevels , aCount);

        for i := 0 to FilterCount - 1
        do begin
          FFilters[ i]            := TSvfChamberlinRelative.Create( Polyphony);
          FFilters[ i].SampleRate := SampleRate;

          for zz := 0 to Polyphony - 1
          do begin
            FFilters[ i].Frequency[ zz] := FrequencyToUnits( 440.0);
            FFilters[ i].Q        [ zz]  := 1.0;
          end;

          FLevels [ i]            := 1.0;
        end;
      finally
        FLocked := False;
      end;
    end;


//  public

    constructor TSvfChladniFilter.Create( aPolyphony: Integer; aSampleRate: TSignal; aMaxVoices : Integer);
    begin
      inherited Create;
      FLocked    := True;
      FPolyphony := aPolyphony;
      FMaxVoices := aMaxVoices;
      SampleRate := aSampleRate;
      RecreateFilter;
      FLocked    := False;
    end;


    destructor  TSvfChladniFilter.Destroy; // override;
    var
      i : Integer;
    begin
      FLocked := True;

      for i := 0 to Length( FFilters) - 1
      do FreeAndNil( FFilters[ i]);

      inherited;
    end;


    procedure   TSvfChladniFilter.Process( aVoice: Integer; const anInput: TSignal; out aValue: TSignal);
    var
      i           : Integer;
      aLowPass    : TSignal; // note : need to use distinct variables for all params
      aBandPass   : TSignal; //        passed to FFilters[ i].Process, or odd stuff
      aBandReject : TSignal; //        will happen (even when the values are not being
      aHighPass   : TSignal; //        used).
    begin
      aValue := 0.0;

      if   not FLocked
      then begin
        for i := 0 to Length( FFilters) - 1
        do begin
          FFilters[ i].Process( aVoice, anInput, aLowPass, aBandPass, aBandReject, aHighPass);

          if   FIs
          then begin
            if   Odd( i)
            then aValue := Normalize( aValue - FLevels[ i] * aBandPass)
            else aValue := Normalize( aValue + FLevels[ i] * aBandPass);
          end
          else aValue := Normalize( aValue + FLevels[ i] * aBandPass);
        end;
      end;
    end;


end.

