unit WaveRecording;

  {$ifdef DEBUG}
    { $define TEST_WAVERECORDER}
  {$endif}

interface

uses

  System.Classes, System.SysUtils,

  KnobsUtils, KnobsConversions, wave_tools;

type

  EWaveRecording       = class( Exception     );
  EWaveInvalidFormat   = class( EWaveRecording);
  EWaveUnhandledFormat = class( EWaveRecording);
  EWaveWriteError      = class( EWaveRecording);


  TWaveRecorder = class
  private
    FData                : TWaveDataSingle;
    FInterpolationMemory : TSignalArray;
    FPreparingRelease    : Boolean;
    FNormalizer          : TSignal;
    FSampleRate          : UInt32;
    FFillCount           : UInt32;
    FInPointer           : UInt32;
    FOutPointer          : UInt32;
    FMaxDuration         : TSignal;
    FMaxDurationStr      : string;
  private
    function    GetSampleCount: UInt32;
    function    GetFillTime   : TSignal;
    function    GetFull       : Boolean;
    procedure   CalcNormalization( UseNormalization: Boolean);
    function    DurationToSampleCount( aDuration   : UInt16): UInt32;
    function    SampleCountToDuration( aSampleCount: UInt32): TSignal;
  public
    constructor Create( aChannelCount, aDuration: UInt16; aSampleRate: Uint32);
    destructor  Destroy;                                                                                       override;
    procedure   Reset;
    procedure   PrepareRelease; // Lock all IO and free all data
    procedure   WriteToFile( const aFileName: string; anAutoNaming, aNormalized: Boolean);
    procedure   WriteAsTextFile( const aFileName: string; anAutoNaming: Boolean; aSignalType: Integer);
    procedure   WriteSamples( const aValue: TSignalArray);
    procedure   ReadSamplesLinear ( aPosition: TSignal; var aResult: TSignalArray);
    procedure   ReadSamplesAllpass( aPosition: TSignal; var aResult: TSignalArray);
  public
    property    SampleCount    : UInt32  read GetSampleCount;
    property    FillTime       : TSignal read GetFillTime;
    property    Full           : Boolean read GetFull;
    property    InPointer      : Uint32  read FInPointer;
    property    MaxDuration    : TSignal read FMaxDuration;
    property    MaxDurationStr : string  read FMaxDurationStr;
  end;




implementation


{ ========
  TWaveRecorder = class
  private
    FData                : TWaveDataSingle;
    FInterpolationMemory : TSignalArray;
    FPreparingRelease    : Boolean;
    FNormalizer          : TSignal;
    FSampleRate          : UInt32;
    FFillCount           : UInt32;
    FInPointer           : UInt32;
    FOutPointer          : UInt32;
    FMaxDuration         : TSignal;
    FMaxDurationStr      : string;
  public
    property    SampleCount    : UInt32  read GetSampleCount;
    property    FillTime       : TSignal read GetFillTime;
    property    Full           : Boolean read GetFull;
    property    InPointer      : Uint32  read FInPointer;
    property    MaxDuration    : TSignal read FMaxDuration;
    property    MaxDurationStr : string  read FMaxDurationStr;
  private
}

    function    TWaveRecorder.GetSampleCount: UInt32;
    begin
      if   Length( FData) = 0
      then Result := 0
      else Result := Length( FData[ 0]);
    end;


    function    TWaveRecorder.GetFillTime: TSignal;
    begin
      Result := SampleCountToDuration( FFillCount);
    end;


    function    TWaveRecorder.GetFull: Boolean;
    begin
      Result := FFillCount >= SampleCount;
    end;


    procedure   TWaveRecorder.CalcNormalization( UseNormalization: Boolean);
    var
      aPeak  : TSignal;
      aValue : TSignal;
      i      : Integer;
      j      : Integer;
      p      : Integer;
    begin
      if   UseNormalization
      then begin
        aPeak := 0;
        p     := FOutPointer;

        for j := 0 to FFillCount - 1
        do begin
          for i := 0 to Length( FData) - 1
          do begin
            aValue := Abs( FData[ i, p]);

            if   aValue > aPeak
            then aPeak := aValue;
          end;

          p := ( p + 1) mod Length( FData[ 0]);
        end;

        if   AlmostZero( aPeak)
        then FNormalizer := 0.0
        else FNormalizer := 0.95 / aPeak;      // With 95% 'true peaks' seem to stay well below 0 dB
      end
      else FNormalizer := 1.0;
    end;


    function    TWaveRecorder.DurationToSampleCount( aDuration: UInt16): UInt32;
    begin
      Result := FSampleRate * aDuration;
    end;


    function    TWaveRecorder.SampleCountToDuration( aSampleCount: UInt32): TSignal;
    begin
      if   FSampleRate <> 0
      then Result := aSampleCount / FSampleRate
      else Result := 0;
    end;


//  public

    constructor TWaveRecorder.Create( aChannelCount, aDuration: UInt16; aSampleRate: Uint32);
    const
      HEAD_SIZE    = 44;     // For a canonical WAVE file .. and that is what we make here
                             // could be calculated .. but I'm lazy ...
      MEMORY_DELTA = $2000;  // Inaccessible constant (MemoryDelta there) in System.Classes .. so .. subject to unnoticed change over releases .. fuckers!
                             // apart from that there is a bug in TMemoryStream where MEMORY_DELTA will be added when capacity
                             // needs to be enlarged .. but .. without checking if the resulting size fits into an Int32 variable .. fuckers!
                             // Soo .. need to be safe and reserve 2 * MEMORY_DELTA ... fuckers!
                             // Then there is an issue with win32 having a total memory limit of 2GiB ...
    var
      i            : Integer;
      Length       : UInt64;
      WaveDataSize : UInt64;
      MaxSize      : UInt64;
      BlockSize    : UInt64;
      aTTime       : TDateTime;
    begin
      inherited Create;
      FNormalizer  := 1.0;
      FSampleRate  := aSampleRate;
      Length       := DurationToSampleCount( aDuration);
      BlockSize    := Uint64( aChannelCount) * UInt64( SizeOf( Single));
      WaveDataSize := UInt64( HEAD_SIZE) + Length * BlockSize;
      MaxSize      := Int32.MaxValue - 2 * MEMORY_DELTA - UInt64( HEAD_SIZE); // Restricted by TMemoryStream implementation

      if   aChannelCount = 0
      then raise EWaveInvalidFormat.Create( 'TWaveRecorder.Create - channel count must be > 0');

      if   aSampleRate = 0
      then raise EWaveInvalidFormat.Create( 'TWaveRecorder.Create - sample rate must be > 0');

      if   Length = 0
      then raise EWaveInvalidFormat.Create( 'TWaveRecorder.Create - size must be > 0');

      if   WaveDataSize > MaxSize
      then Length := MaxSize div ( BlockSize + 1);

    {$ifdef DEBUG}
      WaveDataSize := UInt64( HEAD_SIZE) + Length * BlockSize;

      if   WaveDataSize > MaxSize
      then raise EWaveInvalidFormat.CreateFmt( 'TWaveRecorder.Create - size must be <= %u', [ MaxSize]);
    {$endif}

      FMaxDuration    := Length / aSampleRate;
      aTTime          := FMaxDuration / SecsPerDay;
      FMaxDurationStr := FormatDateTime( 'nn:ss', aTTime);

      SetLength( FData               , aChannelCount);
      SetLength( FInterpolationMemory, aChannelCount);

      for i := 0 to aChannelCount - 1
      do SetLength( FData[ i], Length);

      Reset;
    end;


    destructor  TWaveRecorder.Destroy; // override;
    var
      i : Integer;
    begin
      for i := 0 to Length( FData) - 1
      do SetLength( FData[ i], 0);

      SetLength( FData               , 0);
      SetLength( FInterpolationMemory, 0);
      inherited;
    end;


    procedure   TWaveRecorder.Reset;
    begin
      FFillCount  := 0;
      FInPointer  := 0;
      FOutPointer := 0;
    end;


    procedure   TWaveRecorder.PrepareRelease;
    var
      i : Integer;
    begin
      FPreparingRelease := True;
      Reset;

      for i := 0 to Length( FData) - 1
      do SetLength( FData[ i], 0);

      SetLength( FData               , 0);
      SetLength( FInterpolationMemory, 0);
    end;


    procedure   TWaveRecorder.WriteToFile( const aFileName: string; anAutoNaming, aNormalized: Boolean);
    var
      aWriter      : TRiffReaderWriter;
      aStream      : TBytesStream;
      i            : Integer;
      j            : Integer;
      p            : Integer;
      aSample      : Single;
      anUInt16     : UInt16;
      aStreamSpeed : UInt32;
      SavePath     : string;
      SaveName     : string;
      aSize        : UInt64;
    begin
      if   ( not FPreparingRelease)      // Do not write when we're winding down
      and  ( FillTime > 0)               // Do not write when there was nothing recorded
      then begin
        if   anAutoNaming
        then begin
          SavePath := ExtractFilePath( aFileName);
          SaveName := ExtractFilename( aFileName);
          SaveName :=
            Format(
              '%s%s__%s',
              [
                IncludeTrailingPathDelimiter( SavePath),
                FormatDateTime( 'yyyy_mm_dd__hh_nn_ss__zzz', Now),
                SaveName
              ],
              AppLocale
            );
        end
        else SaveName := aFileName;

        try
          aWriter := TRiffReaderWriter.Create;

          try
            CalcNormalization( aNormalized);
            aStreamSpeed   := FSampleRate * UInt32( Length( FData)) * Uint32( SizeOf( aSample)) * 8;
            aWriter.FourCC := 'WAVE';
            aStream := aWriter.AddChunk( 'fmt ');
            aStream.Seek( 0, soBeginning);
            anUInt16 := 3;                                          // Wave format, floats
            aStream.Write( anUInt16, SizeOf( anUInt16));
            anUInt16 := Length( FData);                             // Channel count
            aStream.Write( anUInt16, SizeOf( anUInt16));
            aStream.Write( FSampleRate, SizeOf( FSampleRate));      // Sample rate
            aStream.Write( aStreamSpeed, SizeOf( aStreamSpeed));    // Stream bit rate
            anUInt16 := Length( FData) * SizeOf( aSample);          // Block align
            aStream.Write( anUInt16, SizeOf( anUInt16));
            anUInt16 := UInt32( SizeOf( aSample)) * 8;              // Bits per sample
            aStream.Write( anUInt16, SizeOf( anUInt16));
            aStream := aWriter.AddChunk( 'data');
            aSize        := UInt64( Length( FData)) * UInt64( FFillCount) * UInt64( SizeOf( Single));
            aStream.Size := aSize;
            aStream.Seek( 0, soBeginning);
            p := FOutPointer;

            for j := 0 to FFillCount - 1
            do begin
              for i := 0 to Length( FData) - 1
              do begin
                aSample := FNormalizer * FData[ i, p];
                aStream.Write( aSample, SizeOf( aSample));
              end;

              p := ( p + 1) mod Length( FData[ 0]);
            end;

            aWriter.SaveToFile( SaveName);
          finally
            aWriter.DisposeOf;
          end;
        except
          on E: Exception
          do raise EWaveWriteError.CreateFmt( 'Wave write error: ', [ E.ToString])
        end;
      end;
    end;


    procedure   TWaveRecorder.WriteAsTextFile( const aFileName: string; anAutoNaming: Boolean; aSignalType: Integer);
    var
      DiskFile     : TextFile;
      SavePath     : string;
      SaveName     : string;
      S            : string;
      F            : string;
      i            : Integer;
      j            : Integer;
      p            : Integer;
      Sample       : Single;
    begin
      if   ( not FPreparingRelease)      // Do not write when we're winding down
      and  ( FillTime > 0)               // Do not write when there was nothing recorded
      then begin
        if   anAutoNaming
        then begin
          SavePath := ExtractFilePath( aFileName);
          SaveName := ExtractFilename( aFileName);
          SaveName :=
            Format(
              '%s%s__%s',
              [
                IncludeTrailingPathDelimiter( SavePath),
                FormatDateTime( 'yyyy_mm_dd__hh_nn_ss__zzz', Now),
                SaveName
              ],
              AppLocale
            );
        end
        else SaveName := aFileName;

        try
          AssignFile( DiskFile, SaveName);
          Rewrite( DiskFile);

          try
            p := FOutPointer;

            for j := 0 to FFillCount - 1
            do begin
              S := '';

              for i := 0 to Length( FData) - 1
              do begin
                Sample := ConvertToSignalType( FData[ i, p], aSignalType);
                F      := Format( '%.6f', [ Sample], AppLocale);

                while Length( F) < 12
                do    F := ' ' + F;

                if   S = ''
                then S :=     Format(   '%s', [ F], AppLocale)
                else S := S + Format( ', %s', [ F], AppLocale);
              end;

              WriteLn( DiskFile, S);
              p := ( p + 1) mod Length( FData[ 0]);
            end;

            if   S <> ''
            then WriteLn( DiskFile, S);
          finally
            CloseFile( DiskFile);
          end;
        except
          on E: Exception
          do raise EWaveWriteError.CreateFmt( 'Data write error: ', [ E.ToString])
        end;
      end;
    end;


    procedure   TWaveRecorder.WriteSamples( const aValue: TSignalArray);
    var
      i : Integer;
    begin
      if   not FPreparingRelease
      then begin
        if   Length( aValue) <> Length( FData)
        then raise EWaveUnhandledFormat.Create( 'TWaveRecorder.AddSamples - number of samples must equal number of channels');

        for i := 0 to Length( FData) - 1
        do FData[ i, FInPointer] := aValue[ i];

        FInPointer := ( FInPointer + 1) mod UInt32( Length( FData[ 0]));

        if   FFillCount < UInt32( Length( FData[ 0]))
        then Inc( FFillCount)
        else FOutPointer := ( FOutPointer + 1) mod UInt32( Length( FData[ 0]));
      end;
    end;


    procedure   TWaveRecorder.ReadSamplesLinear( aPosition: TSignal; var aResult: TSignalArray);
    var
      i : Integer;
    begin
      if   not FPreparingRelease
      then begin
        if   Length( aResult) <> Length( FData)
        then raise EWaveUnhandledFormat.Create( 'TWaveRecorder.ReadSamplesL - number of requested samples must equal number of channels');

        for i := 0 to Length( FData) - 1
        do aResult[ i] := InterpolateWaveLinear( FData, i, aPosition);
      end;
    end;


    procedure   TWaveRecorder.ReadSamplesAllpass( aPosition: TSignal; var aResult: TSignalArray);
    var
      i : Integer;
    begin
      if   not FPreparingRelease
      then begin
        if   Length( aResult) <> Length( FData)
        then raise EWaveUnhandledFormat.Create( 'TWaveRecorder.ReadSamplesA - number of requested samples must equal number of channels');

        for i := 0 to Length( FData) - 1
        do aResult[ i] := InterpolateWaveAllpass( FData, i, aPosition, FInterpolationMemory[ i]);
      end;
    end;



{$ifdef TEST_WAVERECORDER}

  procedure TestWaveRecorder;
  const
    CH_COUNT    = 2;        // Number of channels
    DURATION    = 5 * 60;   // Duration in seconds
    SAMPLE_RATE = 96000;    // Sample rate
  var
    aRecorder : TWaveRecorder;
    aData     : TSignalArray;
    i         : UInt32;
    j         : UInt32;
    L         : UInt32;
  begin
    SetLength( aData, CH_COUNT);
    aRecorder := TWaveRecorder.Create( CH_COUNT, DURATION, SAMPLE_RATE);
    L         := aRecorder.SampleCount;

    try
      for i := 1 to 3 * L div 2
      do begin
        for j := 0 to CH_COUNT - 1
        do aData[ j] := ( 1.0 * i * L) * 1.5 * ( 1.0 - 2.0 * Random);

        aRecorder.WriteSamples( aData);
      end;

      aRecorder.WriteToFile( 'wave_recorder_test.wav', False, False);
    finally
      aRecorder.Disposeof;
    end;
  end;

initialization

  TestWaveRecorder;
  Halt;

{$endif}

end.
