unit WrenLangIf;

{
   COPYRIGHT 2020 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}

interface

uses

  System.SysUtils, System.Generics.Collections,

  WrenLang;



type

  EWrenLangIf          = class( Exception);
  TOnWrenWrite         = procedure( const aSender: TObject; const aText: string) of object;
  TOnWrenError         = procedure( const aSender: TObject; anErrorType: TWrenErrorType; const aModule: string; aLine: Integer; const aMessage: string) of object;
  TOnWrenResolveModule = function ( const aSender: TObject; const anImporter, aName: string): string of object;
  TOnWrenLoadModule    = function ( const aSender: TObject; const aModule: string): string of object;
  TOnWrenBindMethod    = function ( const aSender: TObject; const aModule, aClass, aSignature: string; isStatic: Boolean): TWrenForeignMethodFn of object;
  TOnWrenBindClass     = function ( const aSender: TObject; const aModule, aClass: string): TWrenForeignClassMethods of object;


  TModuleData = record
  public
    TextValue : string;                           // Script text or a file name
    IsFile    : Boolean;                          // when IsFile this is an absolute path, and otherwise it is script text.
  end;
  TModule  = TPair      < string, TModuleData>;   // the string is a module name
  TModules = TDictionary< string, TModuleData>;   // the string is a module name


  TForeignMethodId = record
  public
    ModuleName : string;
    ClassName  : string;
    Signature  : string;
    IsStatic   : Boolean;
  end;
  TForeignMethod  = TPair      < TForeignMethodId, TWrenForeignMethodFn>;
  TForeignMethods = TDictionary< TForeignMethodId, TWrenForeignMethodFn>;


  TForeignClassId = record
  public
    ModuleName : string;
    ClassName  : string;
  end;
  TForeignClassHandlers = record
  public
    Allocator : TWrenForeignMethodFn;
    Finalizer : TWrenFinalizerFn;
  end;
  TForeignClass   = TPair      < TForeignClassId, TForeignClassHandlers>;
  TForeignClasses = TDictionary< TForeignClassId, TForeignClassHandlers>;


  TWrenLangVM = class
  // Class wrapper for WrenLang
  private
    FLineSeparator   : string;
    FConfiguration   : TWrenConfiguration;
    FVirtualMachine  : PWrenVM;
    FModules         : TModules;
    FForeignMethods  : TForeignMethods;
    FForeignClasses  : TForeignClasses;
    FOnWrite         : TOnWrenWrite;
    FOnError         : TOnWrenError;
    FOnResolveModule : TOnWrenResolveModule;
    FOnLoadModule    : TOnWrenLoadModule;
    FOnBindMethod    : TOnWrenBindMethod;
    FOnBindClass     : TOnWrenBindClass;
  private
    function    GetDllVersion: string;
    function    GetSlotCount: Integer;
    procedure   SetSlotCount( aSlotCount: Integer);
    function    GetSlotBool  ( aSlot: Integer): Boolean;
    procedure   SetSlotBool  ( aSlot: Integer; aValue: Boolean);
    function    GetSlotDouble( aSlot: Integer): Double;
    procedure   SetSlotDouble( aSlot: Integer; aValue: Double );
    function    GetSlotBytes ( aSlot: Integer): TBytes;
    procedure   SetSlotBytes ( aSlot: Integer; const aValue: TBytes);
    function    GetSlotString( aSlot: Integer): string;
    procedure   SetSlotString( aSlot: Integer; const aValue: string);
    function    GetSlotHandle( aSlot: Integer): PWrenHandle;
    procedure   SetSlotHandle( aSlot: Integer; aValue: PWrenHandle);
    function    GetSlotType  ( aSlot: Integer): TWrenType;
  private
    procedure   FixConfiguration(
      anInitialHeapSize  : TSize;
      aMinHeapSize       : TSize;
      aHeapGrowthPercent : Integer
    );
  protected
    procedure   WriteHandler       ( aText : string);
    procedure   ErrorHandler       ( anErrorType: TWrenErrorType; const aModule: string; aLine: Integer; aMessage: string);
    function    ModuleResolver     ( const anImporter, aName: string): string;
    function    ModuleLoader       ( aName: string): string;
    function    ForeignMethodBinder( const aModule, aClassName: string; isStatic: Boolean; const aSignature: string): TWrenForeignMethodFn;
    function    ForeignClassBinder ( aModule, aClassName: string): TWrenForeignClassMethods;
  public
    constructor Create(
      aLineSeparator     : string  = #13#10;     // CarriageReturn / LineFeed as  the default line separator
      anInitialHeapSize  : TSize   = 0;          // for the default value (of 10MB)
      aMinHeapSize       : TSize   = 0;          // for the default value (of  1MB)
      aHeapGrowthPercent : Integer = 0           // for the default value (of 50% )
    );
    destructor  Destroy;                                                                                       override;
    procedure   RegisterModuleFile( const aModuleName, aFileName  : string);
    procedure   RegisterModuleText( const aModuleName, aModuleText: string);
    procedure   RegisterForeignMethod( const aModule, aClass, aSignature: string; IsStatic: Boolean; const aMethod: TWrenForeignMethodFn);
    procedure   RegisterForeignClass( const aModule, aClass: string; const anAllocator: TWrenForeignMethodFn; const aFinalizer: TWrenFinalizerFn);
    function    GetSlotForeign( aSlot: Integer): Pointer;
    function    SetSlotNewForeign( aSlot, aClassSlot: Integer; aSize: TSize): Pointer;
    function    Interpret( const aModule, aScript: string): TWrenInterpretResult;
    function    Call( aMethod: PWrenHandle): TWrenInterpretResult;
    procedure   ClearSlot     ( aSlot: Integer);
    procedure   SetSlotNewList( aSlot: Integer);
    procedure   InsertInList( aListSlot, anIndex, anElementSlot: Integer);
    procedure   GetVariable( const aModule, aName: string; aSlot: Integer);
    procedure   ReleaseHandle( const aValue: PWrenHandle);
    function    MakeCallHandle( const aSignature: string): PWrenHandle;
  public
    property    DllVersion                   : string               read GetDllVersion;
    property    SlotCount                    : Integer              read GetSlotCount     write SetSlotCount;
    property    SlotBool   [ aSlot: Integer] : Boolean              read GetSlotBool      write SetSlotBool;
    property    SlotDouble [ aSlot: Integer] : Double               read GetSlotDouble    write SetSlotDouble;
    property    SlotBytes  [ aSlot: Integer] : TBytes               read GetSlotBytes     write SetSlotBytes;
    property    SlotString [ aSlot: Integer] : string               read GetSlotString    write SetSlotString;
    property    SlotHandle [ aSlot: Integer] : PWrenHandle          read GetSlotHandle    write SetSlotHandle;
    property    SlotType   [ aSlot: Integer] : TWrenType            read GetSlotType;
    property    OnWrite                      : TOnWrenWrite         read FOnWrite         write FOnWrite;
    property    OnError                      : TOnWrenError         read FOnError         write FOnError;
    property    OnResolveModule              : TOnWrenResolveModule read FOnResolveModule write FOnResolveModule;
    property    OnLoadModule                 : TOnWrenLoadModule    read FOnLoadModule    write FOnLoadModule;
    property    OnBindMethod                 : TOnWrenBindMethod    read FOnBindMethod    write FOnBindMethod;
    property    OnBindClass                  : TOnWrenBindClass     read FOnBindClass     write FOnBindClass;
  end;


  function WrenErrorTypeToStr( aValue: TWrenErrorType): string;


implementation



var

  GDllPresent    : Boolean = False;
  GDllVersion    : string  = '';
  GDefaultConfig : TWrenConfiguration;


  // ===================================================================================================================
  // Helper functions
  //
  function  WrenErrorTypeToStr( aValue: TWrenErrorType): string;
  // Return a textual representation for the script execution error type
  begin
    case aValue of
      WREN_ERROR_COMPILE     : Result := 'Compilation error';
      WREN_ERROR_RUNTIME     : Result := 'Runtime error';
      WREN_ERROR_STACK_TRACE : Result := 'Stack trace';
      else                     Result := 'Undefined error'
    end;
  end;
  //
  //
  function  InitLanguageVM: TWrenConfiguration;
  // Establish a connection to the WrenLang DLL
  begin
    if   not GDllPresent
    then begin
      try
        GDllVersion := string( PCChar( wrenAPIVersion));
        wrenInitConfiguration( @ Result);
        GDefaultConfig := Result;
        GDllPresent    := True;
      except
        raise EWrenLangIf.Create( 'Could not initialize the Wren language DLL, tried to find: ' + wren_dllpath);
      end;
    end
    else Result := GDefaultConfig;
  end;
  //
  //
  function FindVMObject( vm: PWrenVM; const aSrc: string): TWrenLangVM;
  // For callbacks, obtain the user data from the vm, it is supposed to be a pointer to the invoking TWrenLangVM instance
  begin
    if   Assigned( vm)
    then Result := wrenGetUserData( vm)
    else Result := nil;

    if   not Assigned( Result)
    then raise EWrenLangIf.Create( 'Could not find a Wren languange handling object for the passed vm (from ''' + aSrc + ''')');
  end;
  //
  // ===================================================================================================================

  // ===================================================================================================================
  // WrenLang callbacks
  //
  function  GMemoryAllocator( aMemory: Pointer; aNewSize: TSize): Pointer; cdecl;
  // Memory allocator for the WrenLang, shared by all instances
  begin
    try
      if   Assigned( aMemory)
      then begin
        if   aNewSize = 0
        then begin
          FreeMem( aMemory);
          Result := nil;
        end
        else begin
          ReAllocMem( aMemory, aNewSize);
          Result := aMemory;
        end;
      end
      else Result := AllocMem( aNewSize);
    except
      Result := nil;
    end;
  end;
  //
  //
  function  GModuleResolver( vm: PWrenVM; const anImporter, aName: PCChar): PCChar;                               cdecl;
  var
    S : UTF8String;
  begin
    S      := UTF8String( FindVMObject( vm, 'GModuleResolver').ModuleResolver( string( anImporter), string( aName)));
    Result := PCChar( S);
  end;
  //
  //
  function  GModuleLoader( vm: PWrenVM; aName: PCChar): PCChar;                                                   cdecl;
  var
    S : UTF8String;
  begin
    S      := UTF8String( FindVMObject( vm, 'GModuleLoader').ModuleLoader( string( aName)));
    Result := PCChar( S);
  end;
  //
  //
  function  GForeignMethodBinder( vm: PWrenVM; const aModule, aClassName: PCChar; isStatic: Boolean; const aSignature: PCChar): TWrenForeignMethodFn; cdecl;
  begin
    Result := FindVMObject( vm, 'GForeignMethodBinder').ForeignMethodBinder( string( aModule), string( aClassName), isStatic, string( aSignature));
  end;
  //
  //
  function  GForeignClassBinder( vm: PWrenVM; aModule, aClassName: PCChar): TWrenForeignClassMethods; cdecl;
  begin
    Result := FindVMObject( vm, 'GForeignClassBinder').ForeignClassBinder( string( aModule), string( aClassName));
  end;
  //
  //
  procedure GWriteHandler( vm: PWrenVM; aText : PCChar); cdecl;
  begin
    FindVMObject( vm, 'GWriteHandler').WriteHandler( string( aText));
  end;
  //
  //
  procedure GErrorHandler( vm: PWrenVM; anErrorType: TWrenErrorType; const aModule: PCChar; aLine: Integer; aMessage: PCChar); cdecl;
  begin
    FindVMObject( vm, 'GErrorHandler').ErrorHandler( anErrorType, string( aModule), aLine, string( aMessage));
  end;
  //
  // ===================================================================================================================


  // ===================================================================================================================
  // Class wrapper implementation for WrenLang
  //
  { ========
    TWrenLangVM = class
    private
  }

    function    TWrenLangVM.GetDllVersion: string;
    begin
      Result := GDllVersion;
    end;


    function    TWrenLangVM.GetSlotCount: Integer;
    begin
      Result := wrenGetSlotCount( FVirtualMachine);
    end;


    procedure   TWrenLangVM.SetSlotCount( aSlotCount: Integer);
    begin
      wrenEnsureSlots( FVirtualMachine, aSlotCount);
    end;


    function    TWrenLangVM.GetSlotBool( aSlot: Integer): Boolean;
    begin
      Result := wrenGetSlotBool( FVirtualMachine, aSlot);
    end;


    procedure   TWrenLangVM.SetSlotBool( aSlot: Integer; aValue: Boolean);
    begin
      wrenSetSlotBool( FVirtualMachine, aSlot, aValue);
    end;


    function    TWrenLangVM.GetSlotDouble( aSlot: Integer): Double;
    begin
      Result := wrenGetSlotDouble( FVirtualMachine, aSlot);
    end;


    procedure   TWrenLangVM.SetSlotDouble( aSlot: Integer; aValue: Double);
    begin
      wrenSetSlotDouble( FVirtualMachine, aSlot, aValue);
    end;


    function    TWrenLangVM.GetSlotBytes ( aSlot: Integer): TBytes;
    var
      aLength : Integer;
      Data    : PCChar;
    begin
      Data := wrenGetSlotBytes( FVirtualMachine, aSlot, @ aLength);
      SetLength( Result, aLength);
      Move( Data, Result[ 0], aLength);
    end;


    procedure   TWrenLangVM.SetSlotBytes( aSlot: Integer; const aValue: TBytes);
    begin
      if   Assigned( aValue)
      then wrenSetSlotBytes( FVirtualMachine, aSlot, PCChar( @ aValue[ 0]), Length( aValue));
    end;


    function    TWrenLangVM.GetSlotString( aSlot: Integer): string;
    var
      Data : PCChar;
    begin
      Data   := wrenGetSlotString( FVirtualMachine, aSlot);
      Result := string( Data);
    end;


    procedure   TWrenLangVM.SetSlotString( aSlot: Integer; const aValue: string);
    var
      S : UTF8String;
    begin
      if   Length( aValue) > 0
      then begin
        S := UTF8String( aValue);
        WrenSetSlotString( FVirtualMachine, aSlot, PCChar( S));
      end;
    end;


    function    TWrenLangVM.GetSlotHandle( aSlot: Integer): PWrenHandle;
    begin
      Result := wrenGetSlotHandle( FVirtualMachine, aSlot);
    end;


    procedure   TWrenLangVM.SetSlotHandle( aSlot: Integer; aValue: PWrenHandle);
    begin
      wrenSetSlotHandle( FVirtualMachine, aSlot, aValue);
    end;


    function    TWrenLangVM.GetSlotType( aSlot: Integer): TWrenType;
    begin
      Result := wrenGetSlotType( FVirtualMachine, aSlot);
    end;


//  private

    procedure   TWrenLangVM.FixConfiguration(
      anInitialHeapSize  : TSize;
      aMinHeapSize       : TSize;
      aHeapGrowthPercent : Integer
    );
    begin
      FConfiguration.reallocateFn        := GMemoryAllocator;
      FConfiguration.resolveModuleFn     := GModuleResolver;
      FConfiguration.loadModuleFn        := GModuleLoader;
      FConfiguration.bindForeignMethodFn := GForeignMethodBinder;
      FConfiguration.bindForeignClassFn  := GForeignClassBinder;
      FConfiguration.writeFn             := GWriteHandler;
      FConfiguration.errorFn             := GErrorHandler;
      FConfiguration.initialHeapSize     := anInitialHeapSize;
      FConfiguration.minHeapSize         := aMinHeapSize;
      FConfiguration.heapGrowthPercent   := aHeapGrowthPercent;
      FConfiguration.UserData            := Self;
    end;


//  protected

    procedure   TWrenLangVM.WriteHandler( aText : string);
    begin
      if   Assigned( FOnWrite)
      then FOnWrite( Self, aText)
      else raise EWrenLangIf.Create( 'TWrenLangVM.WriteHandler not implemented');
    end;


    procedure   TWrenLangVM.ErrorHandler( anErrorType: TWrenErrorType; const aModule: string; aLine: Integer; aMessage: string);
    begin
      if   Assigned( FOnError)
      then FOnError( Self, anErrorType, aModule, aLine, aMessage)
      else
        WriteHandler(
            WrenErrorTypeToStr( anErrorType)
          + ''''
          + aMessage
          + ''' in line '
          + IntToStr( aLine)
          + ' in module '''
          + aModule
          + '''.'
        );
    end;


    function    TWrenLangVM.ModuleResolver( const anImporter, aName: string): string;
    begin
      if   Assigned( FOnResolveModule)
      then Result := FOnResolveModule( Self, anImporter, aName)
      else Result := aName;
    end;


    function    TWrenLangVM.ModuleLoader( aName: string): string;
    var
      aData : TModuleData;
      aFile : TextFile;
      L     : string;
    begin
      if   Assigned( FOnLoadModule)
      then Result := FOnLoadModule( Self, aName)    // try external module loader
      else begin                                    // try internal module loader
        if   FModules.TryGetValue( aName, aData)
        then begin
          if  aData.IsFile
          then begin                                // The module is supposed to be a text file on disk
            Result := '';                           // try to load it from file

            if   FileExists( aData.TextValue)       // todo: exclude con:, prn: and such shit?
            then begin
              AssignFile( aFile, aData.TextValue);

              try
                Reset( aFile);

                while not Eof( aFile)
                do begin
                  ReadLn( aFile, L);

                  if   Result = ''
                  then Result := L
                  else Result := Result + FLineSeparator + L;
                end;
              finally
                CloseFile( aFile);
              end;
            end;

            aData.IsFile    := False;               // Do not try to read the file again, but keep the file text in memory as TextValue
            aData.TextValue := Result;
            FModules.AddOrSetValue( aName, aData);
          end
          else Result := aData.TextValue;           // The module is supposed to be stored text
        end
        else Result := '';
      end;
    end;


    function    TWrenLangVM.ForeignMethodBinder( const aModule, aClassName: string; isStatic: Boolean; const aSignature: string): TWrenForeignMethodFn;
    var
      anId    : TForeignMethodId;
      aMethod : TWrenForeignMethodFn;
    begin
      if   Assigned( FOnBindMethod)
      then Result := FOnBindMethod( Self, aModule, aClassName, aSignature, isStatic) // try external bind
      else begin                                                                     // try internal bind
        anId.ModuleName := aModule;
        anId.ClassName  := aClassName;
        anId.Signature  := aSignature;
        anId.IsStatic   := IsStatic;

        if   FForeignMethods.TryGetValue( anId, aMethod)
        then Result := aMethod
        else Result := nil;
      end;
    end;


    function    TWrenLangVM.ForeignClassBinder( aModule, aClassName: string): TWrenForeignClassMethods;
    var
      anId      : TForeignClassId;
      aHandlers : TForeignClassHandlers;
    begin
      if   Assigned( FOnBindClass)
      then Result := FOnBindClass( Self, aModule, aClassName) // try external bind
      else begin                                              // Try internal bind
        anId.ModuleName := aModule;
        anId.ClassName  := aClassName;

        if FForeignClasses.TryGetValue( anId, aHandlers)
        then begin
          Result.allocate := aHandlers.Allocator;
          Result.finalize := aHandlers.Finalizer;
        end
        else begin
          Result.allocate := nil;
          Result.finalize := nil;
        end;
      end;
    end;


//  public

    constructor TWrenLangVM.Create(
      aLineSeparator     : string  = #13#10;     // CarriageReturn / LineFeed as  the default line separator
      anInitialHeapSize  : TSize   = 0;          // for the default value of 10MB
      aMinHeapSize       : TSize   = 0;          // for the default value of  1MB
      aHeapGrowthPercent : Integer = 0           // for the default value of 50%
    );
    begin
      inherited Create;
      FModules        := TModules       .Create;
      FForeignMethods := TForeignMethods.Create;
      FForeignClasses := TForeignClasses.Create;
      FLineSeparator  := aLineSeparator;
      FConfiguration  := InitLanguageVM;
      FixConfiguration( anInitialHeapSize, aMinHeapSize, aHeapGrowthPercent);
      FVirtualMachine := wrenNewVM( @ FConfiguration);
    end;


    destructor  TWrenLangVM.Destroy; //override;
    begin
      wrenFreeVM( FVirtualMachine);
      FreeAndNil( FForeignClasses);
      FreeAndNil( FForeignMethods);
      FreeAndNil( FModules       );
      inherited;
    end;


    procedure   TWrenLangVM.RegisterModuleFile( const aModuleName, aFileName: string);
    var
      aData : TModuleData;
    begin
      aData.TextValue := aFilename;
      aData.IsFile    := True;
      FModules.Add( aModuleName, aData);
    end;


    procedure   TWrenLangVM.RegisterModuleText( const aModuleName, aModuleText: string);
    var
      aData : TModuleData;
    begin
      aData.TextValue := aModuleText;
      aData.IsFile    := False;
      FModules.Add( aModuleName, aData);
    end;


    procedure   TWrenLangVM.RegisterForeignMethod( const aModule, aClass, aSignature: string; IsStatic: Boolean; const aMethod: TWrenForeignMethodFn);
    var
      anId : TForeignMethodId;
    begin
      anId.ModuleName := aModule;
      anId.ClassName  := aClass;
      anId.Signature  := aSignature;
      anId.IsStatic   := IsStatic;
      FForeignMethods.Add( anId, aMethod);
    end;


    procedure   TWrenLangVM.RegisterForeignClass( const aModule, aClass: string; const anAllocator: TWrenForeignMethodFn; const aFinalizer: TWrenFinalizerFn);
    var
      anId      : TForeignClassId;
      aHandlers : TForeignClassHandlers;
    begin
      anId     .ModuleName := aModule;
      anId     .ClassName  := aClass;
      aHandlers.Allocator  := anAllocator;
      aHandlers.Finalizer  := aFinalizer;
      FForeignClasses.Add( anId, aHandlers);
    end;


    function    TWrenLangVM.GetSlotForeign( aSlot: Integer): Pointer;
    begin
      Result := wrenGetSlotForeign( FVirtualMachine, aSlot);
    end;


    function    TWrenLangVM.SetSlotNewForeign( aSlot, aClassSlot: Integer; aSize: TSize): Pointer;
    begin
      Result := wrenSetSlotNewForeign( FVirtualMachine, aSlot, aClassSlot, aSize);
    end;


    function    TWrenLangVM.Interpret( const aModule, aScript: string): TWrenInterpretResult;
    var
      M : UTF8String;
      S : UTF8String;
    begin
      M := UTF8String( aModule);
      S := UTF8String( aScript);
      Result := wrenInterpret( FVirtualMachine, PCChar( M), PCChar( S));
    end;


    function    TWrenLangVM.Call( aMethod: PWrenHandle): TWrenInterpretResult;
    begin
      Result := wrenCall( FVirtualMachine, aMethod);
    end;


    procedure   TWrenLangVM.ClearSlot( aSlot: Integer);
    begin
      wrenSetSlotNull( FVirtualMachine, aSlot);
    end;


    procedure   TWrenLangVM.SetSlotNewList( aSlot: Integer);
    begin
      wrenSetSlotNewList( FVirtualMachine, aSlot);
    end;


    procedure   TWrenLangVM.InsertInList( aListSlot, anIndex, anElementSlot: Integer);
    begin
      wrenInsertInList( FVirtualMachine, aListSlot, anIndex, anElementSlot);
    end;


    procedure   TWrenLangVM.GetVariable( const aModule, aName: string; aSlot: Integer);
    var
      M : UTF8String;
      N : UTF8String;
    begin
      M := Utf8String( aModule);
      N := Utf8String( aName  );
      wrenGetVariable( FVirtualMachine, PCChar( M), PCChar( N), aSlot);
    end;


    procedure   TWrenLangVM.ReleaseHandle( const aValue: PWrenHandle);
    begin
      wrenReleaseHandle( FVirtualMachine, aValue);
    end;


    function    TWrenLangVM.MakeCallHandle( const aSignature: string): PWrenHandle;
    var
      S : UTF8String;
    begin
      S      := UTF8String( aSignature);
      Result := wrenMakeCallHandle( FVirtualMachine, PCChar( S));
    end;


  //
  // ===================================================================================================================

end.
