unit loro_sc;

  // C++ version (c) 2004 Russell Borogove / www.tinygod.com
  // Delphi Pascal Version 2005, Thaddy de Koning / www.thaddy.com
  //
  // Lorenz/Rossler iterative function systems as LFOs
  //

  //
  // This module defines the classes TLorenzOsc and TRosslerOsc - low frequency
  // oscillators suitable for modeling 'analog drift' or other random-but-smooth
  // processes. Both classes have identical APIs - you could unify the interface
  // with virtual functions easily.
  //
  // SetSampleRate:
  // Sets the sample rate at which the Iterate function will be called. Only
  // meaningful in conjunction with the SetFreq function.
  //
  // SetFreq:
  // Sets the fundamental frequency of the oscillator. The Rossler oscillator
  // should exhibit harmonic peaks at multiples of that frequency; the Lorenz
  // oscillator has a linear frequency-amplitude relation, so SetFreq will
  // only control the scale of waveform features in a general way.
  //
  // Iterate:
  // Advances the clock by one sample period and returns the value of the
  // function at the current clock; it should be called once per sample-tick.
  //
  // GetCurrent:
  // Returns the same value returned by the latest call to Iterate. Useful
  // in cases where one generator modulates multiple destinations, for example.
  //
  // GetAlternate:
  // Returns a value separate from the current value but correlated with it;
  // these are the X and Y values used for the well-known "butterfly" plots
  // of the Lorenz and Rossler functions. You can use GetAlternate if you
  // want two separate LFOs which are related in mysterious ways at a low
  // cost - for example, you can fine-tune one audio oscillator with the return
  // from Iterate and another oscillator with the return from GetAlternate.
  //
  // Both the primary and alternate returns are calibrated to a -1.0 to +1.0
  // range in normal usage. The implementation is discrete, though, so if the
  // sample rate is low or the frequency high, it may occasionally jump outside
  // that range -- the user is responsible for clamping if the range is
  // critical.
  //

  // 2015-06-09 : Simplified for use in Wren by Blue Hell
  //
  //   - Iterate replaced by Tick().
  //   - Tick() will produce all three output values at the same time.
  //   - GetCurrent was removed.
  //   - GetAlternate was removed.
  //   - The Z value is returned as the 3rd output.
  //   - Added oversampling to make the algorithms stable for higher frequencies.
  //     should be able to generate 12k5 Hz at a sample rate of 44k1 s/s now.
  //     For Lorenz FDT should get not larger than about 0.02, for Rssler the
  //     limit is about 0.08. Added a couple of debug checks for those bounds.
  //   - Factored out a common ancestor for both.
  //   - Unified outputscaling to be the same for all three outputs.
  //   - Replaced some variables used as constants by constants.
  //   - Renamed stuff.
  //   - Added some code conditionally compiled for debug versions only to make
  //     it easier to track issues with the code.


interface

uses

  SysUtils, Math,

  KnobsUtils, KnobsConversions;


type

  TAttractor = class
  // Base class for attractor type classes.
  protected
  type TAttractorState = record
      FFreq : Double;
      FDT   : Double;
      FDX   : Double;
      FDY   : Double;
      FDZ   : Double;
      FX    : Double;
      FY    : Double;
      FZ    : Double;
    end;
    PAttractorState = ^ TAttractorState;
  protected
    FRate      : Double;
    FPolyphony : Integer;
    FState     : array of TAttractorState;
  protected
    procedure   Initialize( aVoice: Integer);                                                                   virtual;
  public
    procedure   SetSampleRate( aRate: Double);                                                        virtual; abstract;
    procedure   SetFreq      ( aVoice: Integer; aFreq: Double);                                       virtual; abstract;
    constructor Create( aPolyphony: Integer);
    procedure   Reset;                                                                                overload; virtual;
    procedure   Reset( aVoice: Integer);                                                              overload; virtual;
  public
    property    Polyphony: Integer read FPolyphony;
  end;


  TLorenzOsc = class( TAttractor)
  // Lorenz system - very broad spectrum noise function with amplitude
  // decreasing with increasing frequency, but tight short-term correlation.
  //
  // The scale of waveform features will change somewhat with the set frequency
  // and sample rate, but not drastically - it's fairly fractal. In particular,
  // there will not be substantial spectral peaks at multiples of the frequency
  // selected by SetFreq.
  private
    procedure   FixDT( aVoice: Integer);
  public
    procedure   SetSampleRate( aRate: Double);                                                                 override;
    procedure   SetFreq      ( aVoice: Integer; aFreq: Double);                                                override;
    procedure   Tick( aVoice: Integer; const Drive: Double; out Out1, Out2, Out3: Double);
  end;


  TRosslerOsc = class( TAttractor)
  // Rssler system - broad spectrum noise function with amplitude
  // decreasing with increasing frequency, and distinct harmonic peaks. The
  // peaks should occur at harmonics of the frequency set by SetFreq.
  private
    procedure   FixDT( aVoice: Integer);
  public
    procedure   SetSampleRate( aRate: Double);                                                                 override;
    procedure   SetFreq      ( aVoice: Integer; aFreq: Double);                                                override;
    procedure   Tick( aVoice: Integer; const Drive: Double; out Out1, Out2, Out3: Double);
  end;


  TVanDerPolOsc = class( TAttractor)
  private
    FOverSampleRate : Integer;
  private
    procedure   FixDT( aVoice: Integer);
  public
    constructor Create( aPolyphony: Integer; OversampleRate: Integer);
    procedure   SetSampleRate( aRate: Double);                                                                 override;
    procedure   SetFreq      ( aVoice: Integer; aFreq: Double);                                                override;
    procedure   Tick( aVoice: Integer; const drive, mu: Double; out Out1, Out2: Double);
  end;



implementation



const

  LORENZ_OVERSAMPLE  = 16;
  LORENZ_SCALE       =  1.0 / 15.0;
  LORENZ_A           = 10.0;
  LORENZ_B           = 28.0;
  LORENZ_C           =  8.0 / 3.0;

  ROSSLER_OVERSAMPLE = 16;
  ROSSLER_SCALE      =  1.0 / 15.0;
  ROSSLER_A          =  0.15;
  ROSSLER_B          =  0.20;
  ROSSLER_C          = 10.0;

{$IFDEF DEBUG}
  procedure Breakpoint;
  begin
  end;
{$ENDIF}


{ ========
  TAttractor = class
  protected
  type TAttractorState = record
      FFreq : Double;
      FDT   : Double;
      FDX   : Double;
      FDY   : Double;
      FDZ   : Double;
      FX    : Double;
      FY    : Double;
      FZ    : Double;
    end;
    PAttractorState = ^ TAttractorState;
  protected
    FRate      : Double;
    FPolyphony : Integer;
    FState     : array of TAttractorState;
  public
    property    Polyphony: Integer read FPolyphony;
  protected
}

    procedure   TAttractor.Initialize( aVoice: Integer); // virtual;
    var
      State : PAttractorState;
    begin
      State := @ FState[ aVoice];
      State^.FDX := 0;
      State^.FDY := 0;
      State^.FDZ := 0;
      State^.FX  := 1;
      State^.FY  := 1;
      State^.FZ  := 1;
    end;

//  public

    constructor TAttractor.Create( aPolyphony: Integer);
    var
      zz : Integer;
    begin
      FPolyphony := aPolyphony;
      SetLength( FState, Polyphony);
      SetSampleRate( 44100.0);

      for zz := 0 to Polyphony - 1
      do begin
        FState[ zz].FFreq := 440;
        SetFreq( zz, 440.0);
        Initialize( zz);
      end;
    end;

    procedure   TAttractor.Reset; // virtual; overload;
    var
      zz : Integer;
    begin
      for zz := 0 to Polyphony - 1
      do Reset( zz);
    end;


    procedure   TAttractor.Reset( aVoice: Integer); // virtual; overload;
    begin
      Initialize( aVoice)
    end;


{ ========
  TLorenzOsc = class( TAttractor)
  private
}

    procedure   TLorenzOsc.FixDT( aVoice: Integer);
    var
      State : PAttractorState;
    begin
      State := @ FState[ aVoice];
      State^.FDT := State^.FFreq / ( FRate * LORENZ_OVERSAMPLE);
    {$IFDEF DEBUG}
      if Abs( State^.FDT) > 0.02
      then Breakpoint;
    {$ENDIF}
    end;

//  public

    procedure   TLorenzOsc.SetSampleRate( aRate: Double); // override;
    var
      zz : Integer;
    begin
      FRate := aRate;

      for zz := 0 to Polyphony - 1
      do FixDT( zz);
    end;

    procedure   TLorenzOsc.SetFreq( aVoice: Integer; aFreq: Double); // override;
    begin
      FState[ aVoice].FFreq := aFreq;
      FixDT( aVoice);
    end;

    procedure   TLorenzOsc.Tick( aVoice: Integer; const Drive: Double; out Out1, Out2, Out3: Double);
    var
      State : PAttractorState;
      i : Integer;
    begin
      State := @ FState[ aVoice];
      for i := 1 to LORENZ_OVERSAMPLE
      do begin
        State^.FDX := LORENZ_A * ( State^.FY - State^.FX);
        State^.FDY := State^.FX * ( LORENZ_B - State^.FZ) - State^.FY + Drive;
        State^.FDZ := State^.FX * State^.FY - LORENZ_C * State^.FZ;

        State^.FX := State^.FX + State^.FDX * State^.FDT;
        State^.FY := State^.FY + State^.FDY * State^.FDT;
        State^.FZ := State^.FZ + State^.FDZ * State^.FDT;
      end;

      Out1 := State^.FX * LORENZ_SCALE;
      Out2 := State^.FY * LORENZ_SCALE;
      Out3 := State^.FZ * LORENZ_SCALE;
    end;


{ ========
  TRosslerOsc = class( TAttractor)
  private
}

    procedure   TRosslerOsc.FixDT( aVoice: Integer);
    var
      State : PAttractorState;
    begin
      State := @ FState[ aVoice];
      State^.FDT  := 2.91 * State^.FFreq / ( FRate * ROSSLER_OVERSAMPLE);
    {$IFDEF DEBUG}
      if Abs( State^.FDT) > 0.08
      then Breakpoint;
    {$ENDIF}
    end;

//  public

    procedure   TRosslerOsc.SetSampleRate( aRate: Double); // override;
    var
      zz : Integer;
    begin
      FRate := aRate;

      for zz := 0 to Polyphony - 1
      do FixDT( zz);
    end;

    procedure   TRosslerOsc.SetFreq( aVoice: Integer; aFreq: Double); // override;
    begin
      FState[ aVoice].FFreq := aFreq;
      FixDT( aVoice);
    end;

    procedure   TRosslerOsc.Tick( aVoice: Integer; const Drive: Double; out Out1, Out2, Out3: Double);
    var
      State : PAttractorState;
      i : Integer;
    begin
      State := @ FState[ aVoice];
      for i := 1 to ROSSLER_OVERSAMPLE
      do begin
        State^.FDX := - State^.FY- State^.FZ;
        State^.FDY := State^.FX + ROSSLER_A * State^.FY;
        State^.FDZ := ROSSLER_B + State^.FZ * ( State^.FX - ROSSLER_C + Drive / 4);

        State^.FX := State^.FX + State^.FDX * State^.FDT;
        State^.FY := State^.FY + State^.FDY * State^.FDT;
        State^.FZ := State^.FZ + State^.FDZ * State^.FDT;
      end;

      Out1 := State^.FX * ROSSLER_SCALE;
      Out2 := State^.FY * ROSSLER_SCALE;
      Out3 := State^.FZ * ROSSLER_SCALE;
    end;


{ ========
  TVanDerPolOsc = class( TAttractor)
  private
    FOverSample : Integer;
  private
}
    procedure   TVanDerPolOsc.FixDT( aVoice: Integer);
    var
      State : PAttractorState;
    begin
      State := @ FState[ aVoice];
      State^.FDT  := TWO_PI * State^.FFreq / ( FRate * FOverSampleRate);
    {$IFDEF DEBUG}
      if Abs( State^.FDT) > 0.08
      then Breakpoint;
    {$ENDIF}
    end;


//  public

    constructor TVanDerPolOsc.Create( aPolyphony: Integer; OversampleRate: Integer);
    begin
      FOverSampleRate := OversampleRate;
      inherited Create( aPolyphony);
    end;


    procedure   TVanDerPolOsc.SetSampleRate( aRate: Double); // override;
    var
      zz : Integer;
    begin
      FRate := aRate;

      for zz := 0 to Polyphony - 1
      do FixDT( zz);
    end;


    procedure   TVanDerPolOsc.SetFreq( aVoice: Integer; aFreq: Double); // override;
    begin
      FState[ aVoice].FFreq := aFreq;
      FixDT( aVoice);
    end;


    procedure   TVanDerPolOsc.Tick( aVoice: Integer; const drive, mu: Double; out Out1, Out2: Double);
    var
      State : PAttractorState;
      i     : Integer;
      m     : Double;
    begin
      State := @ FState[ aVoice];
      m     := Clip( mu, 0.01, 10.0);

      for i := 1 to FOverSampleRate
      do begin
        State^.FDX := Clip( Normalize( State^.FX + State^.FY * State^.FDT), -250, 250);
        State^.FDY := Clip( Normalize( State^.FY + ( drive + m * ( 1.0 - State^.FX * State^.FX) * State^.FY - State^.FX) * State^.FDT), -250, 250);
        State^.FX  := State^.FDX;
        State^.FY  := State^.FDY;
      end;

      Out1 := State^.FX / ( 4 + m);
      Out2 := State^.FY / ( 4 + m);
    end;


end.
