unit sieves;

{

   COPYRIGHT 2019 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}


// {$define PerformModuleTests}       // Comment this line out to not compile and run the unit test function
   {$define UseKnobsUtils     }       // Comment this line out to remove the dependencies on KnobsUtils

interface

uses

  System.SysUtils, System.Math

{$ifdef UseKnobsUtils}
  , KnobsUtils, KnobsConversions
{$endif}
  ;


type

  TSieveSet      = array of Boolean;
  TSieveSequence = array of Integer;

{$ifdef UseKnobsUtils}
  TSieveSignals  = TSignalArray;
{$else}
  TSignal        = Double;
  TSieveSignals  = array of Double;

var

  AppLocale: TFormatSettings;

type
{$endif}

  TSieveModule = class
  private
    FModulo            : Integer;
    FResidue           : Integer;
    FGeneratedSequence : TSieveSequence;
    FGeneratedSignals  : TSieveSignals;
    FIntervals         : TSieveSequence;
    FIntervalSignals   : TSieveSignals;
    FSimplified        : string;
    FValueCount        : Integer;
  public
    constructor Create    ( aModulo, aResidue: Integer);
    destructor  Destroy;                                                                                       override;
    function    Union     ( const aSieve: TSieveModule): TSieveModule;                                          virtual;
    function    Intersect ( const aSieve: TSieveModule): TSieveModule;                                          virtual;
    function    Difference( const aSieve: TSieveModule): TSieveModule;                                          virtual;
    function    Complement: TSieveModule;                                                                       virtual;
    function    HasMember( aValue: Integer): Boolean;                                                           virtual;
    function    Dump: string;                                                                                   virtual;
    function    Generate               ( aLimit: Integer): TSieveSequence;                                      virtual;
    function    GenerateSignals        ( aLimit: Integer): TSieveSignals;                                       virtual;
    function    GenerateIntervals      ( aLimit: Integer): TSieveSequence;                                      virtual;
    function    GenerateIntervalSignals( aLimit: Integer): TSieveSignals;                                       virtual;
    procedure   GenerateAll            ( aLimit: Integer);
  public
    property    Modulo            : Integer        read FModulo;
    property    Residue           : Integer        read FResidue;
    property    GeneratedSequence : TSieveSequence read FGeneratedSequence;
    property    GeneratedSignals  : TSieveSignals  read FGeneratedSignals;
    property    Intervals         : TSieveSequence read FIntervals;
    property    IntervalSignals   : TSieveSignals  read FIntervalSignals;
    property    Simplified        : string         read FSimplified;
    property    ValueCount        : Integer        read FValueCount;
  end;


  TSieveLiteral = class( TSieveModule)
  private
    FData : TSieveSet;
  public
    constructor Create( aModulo: Integer; const aData: TSieveSequence);
    function    HasMember( aValue: Integer): Boolean;                                                          override;
    function    Dump: string;                                                                                  override;
  end;


  TSieveExtendable = class( TSieveLiteral)
  protected
    procedure   ExtendTo( aSize: Integer);                                                            virtual; abstract;
  public
    constructor Create;
    function    HasMember( aValue: Integer): Boolean;                                                          override;
  end;


  TSievePrime = class( TSieveExtendable)
  protected
    procedure   ExtendTo( aSize: Integer);                                                                     override;
  public
    function    Dump: string;                                                                                  override;
  end;


  TSieveFibonacci = class( TSieveExtendable)
  protected
    procedure   ExtendTo( aSize: Integer);                                                                     override;
  public
    function    Dump: string;                                                                                  override;
  end;


  TSievePreExtendable = class( TSieveExtendable)
  private
    FSequence    : TSieveSequence;
    FLastElement : Integer;
  protected
    function    GetCount: Integer;
    procedure   SetCount( aValue: Integer);
    function    GetElement( anIndex: Integer): Integer;
    procedure   AddItem ( aValue: Integer);
    procedure   AddNextElement;
    function    CreateNextElement: Integer;                                                                     virtual;
    procedure   Initialize;                                                                                     virtual;
    procedure   ExtendTo( aSize: Integer);                                                                     override;
  public
    function    Generate               ( aLimit: Integer): TSieveSequence;                                     override;
    function    GenerateSignals        ( aLimit: Integer): TSieveSignals;                                      override;
    function    GenerateIntervals      ( aLimit: Integer): TSieveSequence;                                     override;
    function    GenerateIntervalSignals( aLimit: Integer): TSieveSignals;                                      override;
  public
    property    Count                     : Integer read GetCount     write SetCount;
    property    LastElement               : Integer read FLastElement;
    property    Element[ anIndex: Integer]: Integer read GetElement;
  end;


  TSieveHanoi = class( TSievePreExtendable)
  protected
    function    CreateNextElement: Integer;                                                                    override;
    procedure   Initialize;                                                                                    override;
  public
    function    Dump: string;                                                                                  override;
  end;


  TSieveHanoiPegs = class( TSieveExtendable)
  var
    FLatestDepth : Integer;
    FOddSwap     : Boolean;
    FSequence    : TSieveSequence;
  private
    procedure   AddSeq( aValue: Integer);
  protected
    procedure   StorePeg( aPeg: Integer);
    procedure   Tower( aDepth, zero, one, two: Integer);
    procedure   SolveTower( aDepth: Integer);
    procedure   ExtendTo( aSize: Integer);                                                                     override;
  public
    constructor Create( OddSwap: Boolean);
    function    Dump: string;                                                                                  override;
    function    Generate               ( aLimit: Integer): TSieveSequence;                                     override;
    function    GenerateSignals        ( aLimit: Integer): TSieveSignals;                                      override;
    function    GenerateIntervals      ( aLimit: Integer): TSieveSequence;                                     override;
    function    GenerateIntervalSignals( aLimit: Integer): TSieveSignals;                                      override;
  end;


  TSieveHanoiPegs1 = class( TSieveHanoiPegs)
  public
    constructor Create;
    function    Dump: string;                                                                                  override;
  end;


  TSieveHanoiPegs2 = class( TSieveHanoiPegs)
  public
    constructor Create;
    function    Dump: string;                                                                                  override;
  end;


  TSieveHofstadterQ = class( TSievePreExtendable)
  protected
    function    CreateNextElement: Integer;                                                                    override;
    procedure   Initialize;                                                                                    override;
  public
    function    Dump: string;                                                                                  override;
  end;


  TSieveMetaFib = class( TSievePreExtendable)
  private
    FOrder : Integer;
    FAN    : TSieveSequence;
  private
    procedure   AddAN( aValue: Integer);
  protected
    function    CreateNextElement: Integer;                                                                    override;
    procedure   Initialize;                                                                                    override;
  public
    constructor Create( anOrder: Integer);
    function    Dump: string;                                                                                  override;
    function    DumpAN: string;
  public
    property    Order: Integer read FOrder;
  end;


  TSieveRandom = class( TSieveExtendable)
  private
    FChance : TSignal;
  protected
    procedure   ExtendTo( aSize: Integer);                                                                     override;
  public
    constructor Create( aChance: TSignal);
    function    Dump: string;                                                                                  override;
  end;


  TSieveUnary = class( TSieveModule)
  private
    FSieve : TSieveModule;
  public
    constructor Create( const aSieve: TSieveModule);
    destructor  Destroy;                                                                                       override;
  public
    property    Sieve : TSieveModule read FSieve;
  end;


  TSieveBinary = class( TSieveModule)
  private
    FLeft  : TSieveModule;
    FRight : TSieveModule;
  public
    constructor Create( const aLeft, aRight: TSieveModule);
    destructor  Destroy;                                                                                       override;
  public
    property    Left  : TSieveModule read FLeft;
    property    Right : TSieveModule read FRight;
  end;


  TSieveUnion = class( TSieveBinary)
  public
    function    HasMember( aValue: Integer): Boolean;                                                          override;
    function    Dump: string;                                                                                  override;
  end;


  TSieveIntersection = class( TSieveBinary)
  public
    function    HasMember( aValue: Integer): Boolean;                                                          override;
    function    Dump: string;                                                                                  override;
  end;


  TSieveDifference = class( TSieveBinary)
  public
    function    HasMember( aValue: Integer): Boolean;                                                          override;
    function    Dump: string;                                                                                  override;
  end;


  TSieveComplement = class( TSieveUnary)
  public
    function    HasMember( aValue: Integer): Boolean;                                                          override;
    function    Dump: string;                                                                                  override;
  end;


  procedure SequenceAddNumber   ( var   aSeq: TSieveSequence; aValue: Integer);
  function  SieveSequenceToStr  ( const aSeq: TSieveSequence): string;
  function  StrToSieve          ( var ErrorText: string; const S: string; aPeriod: Integer): TSieveModule;
  function  SieveSignalsToStr   ( const aSeq: TSieveSignals ): string;
  function  SieveSequenceToSieve( const aSeq: TSieveSequence; aPeriod: Integer): TSieveModule;



implementation



{$ifndef UseKnobsUtils}

  function  MathIntMod( X, Y: Integer): Integer; inline; // overload;
  begin
    Result := (( X mod Y) + Y) mod Y;
  end;


  function  IsPrime( aValue: Cardinal): Boolean;
  var
    X       : Cardinal;
    Largest : Cardinal;
  begin
    if   aValue <= 0                               // <= 0 -> not a prime
    then Result := False
    else if   aValue < 4                           // 1, 2 and 3 are primes
    then Result := True
    else begin
      if   ( aValue mod 2) = 0                     // skip all even values to make it a tad faster
      then Result := False
      else begin
        Largest := Trunc( Sqrt( 1.0 * aValue));    // Only check up to Sqrt( aValue)
        Result  := True;                           // Assume prime
        X       := 3;                              // First divisor to try is 3

        while ( X <= Largest) and Result
        do begin
          Inc( X, 2);                              // Go to next odd value
          Result := ( aValue mod X) <> 0;          // Get false when a divisor found, so not prime then.
        end;
      end;
    end;
  end;


  function  IsPerfectSquare( aValue: Integer): Boolean;
  var
    s : Integer;
  begin
    s      := Round( Sqrt( aValue));
    Result := s * s = aValue;
  end;


  function  IsFibonacci( aValue: Integer): Boolean;
  var
    s : Integer;
  begin
    s      := 5 * aValue * aValue;
    Result := IsPerfectSquare( s + 4) or IsPerfectSquare( s - 4);
  end;


{$endif}


  function  RangeMap( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    Result := anOutLow + (( aValue - anInLow) / ( anInHigh - anInLow)) * ( anOutHigh - anOutLow);
  end;


  procedure SequenceAddNumber( var aSeq: TSieveSequence; aValue: Integer);
  begin
    SetLength( aSeq, Length( aSeq) + 1);
    aSeq[ Length( aSeq) - 1 ] := aValue;
  end;


  function SieveSequenceToStr( const aSeq: TSieveSequence): string;
  var
    i : Integer;
    s : Boolean;
  begin
    s      := false;
    Result := '{';

    for i := Low( aSeq) to High( aSeq)
    do begin
      if   s
      then Result := Result + Format( ', %d', [ aSeq[ i]], AppLocale)
      else begin
        s         := True;
        Result    := Result + Format( ' %d' , [ aSeq[ i]], AppLocale);
      end;
    end;

    Result := Result + '}';
  end;


  function SieveSignalsToStr( const aSeq: TSieveSignals): string;
  var
    i : Integer;
    s : Boolean;
  begin
    s      := false;
    Result := '{';

    for i := Low( aSeq) to High( aSeq)
    do begin
      if   s
      then Result := Result + Format( ', %g', [ aSeq[ i]], AppLocale)
      else begin
        s         := True;
        Result    := Result + Format( ' %g' , [ aSeq[ i]], AppLocale);
      end;
    end;

    Result := Result + '}';
  end;


  function  SieveSequenceToSieve( const aSeq: TSieveSequence; aPeriod: Integer): TSieveModule;
  var
    Covered     : TSieveSet;
    Points      : TSieveSet;
    Modules     : TSieveSequence;
    Residues    : TSieveSequence;
    NewModule   : TSieveModule;
    aPoint      : Integer;
    aModulo     : Integer;
    i           : Integer;
    p           : Integer;
    Satisfied   : Boolean;
    NewPoints   : Boolean;
    CoverCount  : Integer;
  begin
    Result := nil;
    SetLength( Covered , aPeriod);
    SetLength( Points  , aPeriod);
    SetLength( Modules , 0      );
    SetLength( Residues, 0      );
    CoverCount := 0;
    p          := 0;

    for i := 0 to aPeriod - 1
    do begin
      Covered[ i] := False;

      if   ( p < Length( aSeq  ))
      and  ( i = aSeq[ p])
      then begin
        Points[ i] := True;
        Inc( p);
      end
      else Points[ i] := False;
    end;

    for aPoint := 0 to aPeriod - 1
    do begin
      aModulo := 2;

      if   Points[ aPoint]
      then begin
        while aModulo < aPeriod
        do begin
          i         := aPoint;
          Satisfied := True;

          while i < aPeriod
          do begin
            if   not Points[ i]
            then begin
              Satisfied := False;
              Break;                        // From while i < aPeriod
            end
            else i := i + aModulo;
          end;

          if   Satisfied
          then begin
            i         := aPoint;
            NewPoints := False;

            while i < aPeriod
            do begin
              if   not Covered[ i]
              then begin
                NewPoints   := True;
                Covered[ i] := True;
                Inc( CoverCount);
              end;

              i := i + aModulo;
            end;

            if   NewPoints
            then begin
              SequenceAddNumber( Modules , aModulo);
              SequenceAddNumber( Residues, aPoint );
            end;
          end;

          if   CoverCount >= Length( aSeq)  // All points covered
          then Break                        // From for aPoint := 0 to Length( Points) - 1
          else Inc( aModulo);
        end;
      end;

      if   CoverCount >= Length( aSeq)      // All points covered
      then Break;                           // From for aPoint := 0 to Length( Points) - 1
    end;

    for i := 0 to Length( Modules) - 1
    do begin
      NewModule := TSieveModule.Create( Modules[ i], Residues[ i]);

      if   Assigned( NewModule)
      then begin
        if   Assigned( Result)
        then Result := Result.Union( NewModule)
        else Result :=               NewModule;
      end;
    end;
  end;


  // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // Sieve parser

  (*
    Sieve   = Term                      // Single term
            | Term '+' Term             // Union of two terms
            | Term '*' Term             // Intersection of two terms
            | Term '^' Term             // Symmetric aifference of two terms
            ;

    Term    = '[' Module  ']'           // Modulo residue sieve
            | '{' Literal '}'           // Integer sequence
            | '(' Sieve   ')'           // parenthesis for grouping
            | '!' Term                  // Inversion of a Term
                                        // all below of form identifier = [a-zA-Z][a-zA-Z0-9]*
            | 'prime'                   // 2,3,5,7,11..
            | 'fibonacci'               // 1,1,2,3,5,8..
            | 'hanoi'                   // 0,1,3,4,7,8.. with the diff output being the ruler function
            | 'hanoipegs1'              // {0,1}, {0,1,3,5}, {0,1,3,5,6,6,7,8}  non alternating
            | 'hanoipegs2'              // {0,2}, {0,1,3,5}, {0,2,3,4,6,6,8,10} alternating
            | 'hofstadterq'             // chaotic
            | 'metafib' '(' num   ')'   // genralization of hanoi, num being the order
            | 'random'  '(' float ')'   // random numbers, float being the chance param
            ;

    Module  = num ',' num               // modulo, residue
            ;

    Literal = Nums '|' num              // { ... | residue}
            ;

    Nums    =  num                      // Comma separated integer sequence
            |  num ',' Nums
            ;

    float   = num                       // floating point number, allows both comma and dot
            | '.' num
            | ',' num
            | num '.' num
            | num ',' num
            ;

    num     = [ '0' .. '9'] *           // Integer number
            ;
  *)

type

  TParseContext = class
  private
    FParseText       : string;
    FParseIndex      : Integer;
    FLineIndex       : Integer;
    FCharIndex       : Integer;
    FLatestCharIndex : Integer;
    FErrorText       : string;
  public
    constructor Create( const aParseText: string);
    procedure   AddError( const aMsg: string);
    procedure   AddErrorFmt( const aFmt: string; const anArgs: array of const);
    function    ParseChar: Char;
    procedure   Unparse( anAmount: Integer);
  public
    property    ErrorText : string read FErrorText;
  end;


{ ========
  TParseContext = class
  private
    FParseText       : string;
    FParseIndex      : Integer;
    FLineIndex       : Integer;
    FCharIndex       : Integer;
    FLatestCharIndex : Integer;
    FErrorText       : string;
  public
    property    ErrorText : string read FErrorText;
  public
}

    constructor TParseContext.Create( const aParseText: string);
    begin
      FParseText        := aParseText;
      FParseIndex       := Low( FParseText);
      FLineIndex        := 1;
      FCharIndex        := 1;
      FLatestCharIndex  := 1;
      FErrorText        := '';
    end;


    procedure   TParseContext.AddError( const aMsg: string);
    begin
      if   FErrorText = ''
      then FErrorText := Format(      '%s @ (%d, %d)', [             aMsg, FLineIndex, FCharIndex], AppLocale)
      else FErrorText := Format( '%s # %s @ (%d, %d)', [ FErrorText, aMsg, FLineIndex, FCharIndex], AppLocale);
    end;


    procedure   TParseContext.AddErrorFmt( const aFmt: string; const anArgs: array of const);
    begin
      AddError( Format( aFmt, anArgs));
    end;


    function    TParseContext.ParseChar: Char;
    begin
      while ( FParseIndex <= High( FParseText))
      and   CharInSet( FParseText[ FParseIndex], [ ' ', ^I, ^M, ^J])
      do begin
        if   FParseText[ FParseIndex] = ^M
        then begin
          Inc( FLIneIndex);
          FLatestCharIndex := FCharIndex;
          FCharIndex       := 1;
        end
        else Inc( FCharIndex);

        Inc( FParseIndex);
      end;

      if   FParseIndex > High( FParseText)
      then Result := Char( -1)
      else begin
        Result      := FparseText[ FParseIndex];
        FParseIndex := FParseIndex + 1;
      end;
    end;


    procedure   TParseContext.Unparse( anAmount: Integer);
    begin
      while ( FParseIndex >= Low( FParseText))
      and   ( anAmount    >  0               )
      do begin
        Dec( FParseIndex);
        Dec( anAmount   );

        if   FCharIndex > 1
        then Dec( FCharIndex);

        if   FParseText[ FParseIndex] = ^M
        then begin
          Dec( FLineIndex);
          FCharIndex := FLatestCharIndex;
        end;
      end;
    end;



  function ParseSieve( const aContext: TParseContext): TSieveModule; forward;


  function ParseIdentifier( const aContext: TParseContext): string;
  var
    c : Char;
  begin                                                // identifier = [ 'a' .. 'z', 'A' .. 'Z'] [ 'a' .. 'z', 'A' .. 'Z', '0' .. '9']*
    Result := '';
    c      := aContext.ParseChar;

    if   CharInSet( c, [ 'a' .. 'z', 'A' .. 'Z' ])
    then begin
      while CharInset( c, [ 'a' .. 'z', 'A' .. 'Z' , '0' .. '9'])
      do begin
        Result := Result + c;
        c      := aContext.ParseChar;
      end;
    end;

    aContext.Unparse( 1);                              // We rad past the num .. so backup one
  end;


  function ParseNum( const aContext: TParseContext; var aValue: Integer): Boolean;
  var
    t : Char;
  begin                                                // num = [ 0 .. 9] *
    Result := False;
    t      := aContext.ParseChar;
    aValue := 0;

    while CharInSet( t, [ '0' .. '9'])
    do begin
      aValue := aValue * 10 + ( Ord( t) - Ord( '0'));
      t      := aContext.ParseChar;
      Result := True;
    end;

    aContext.Unparse( 1);                              // We rad past the num .. so backup one

    if   not Result
    then aContext.AddError( 'number expected');
  end;


  function ParseFloat( const aContext: TParseContext; var F: TSignal): Boolean;
  var
    N : Integer;
    M : Integer;
  begin                                                // float = num | '.' num | num '.' num ;
    Result := False;
    F      := 0;

    if   CharInSet( aContext.ParseChar, [ '.', ','])
    then begin                                         // float = '.' num | ',' num
      if   ParseNum( aContext, N)
      then begin
        F := N;

        while N <> 0
        do begin
          F := F  /  10.0;
          N := N div 10;
        end;

        Result := True;
      end
    end
    else begin                                         // float = num | num '.' num | num ',' num
      aContext.Unparse( 1);

      if   ParseNum( aContext, M)
      then begin
        if   CharInSet( aContext.ParseChar, [ '.', ','])
        then begin                                     // float = num '.' num | num ',' num
          if   ParseNum( aContext, N)
          then begin
            F := N;

            while N <> 0
            do begin
              F := F  /  10.0;
              N := N div 10;
            end;

            F      := F + M;
            Result := True;
          end
          else aContext.AddError( 'float expected, no number seen after the dot or comma');
        end
        else begin                                     // float = num
          aContext.Unparse( 1);
          F      := M;
          Result := True;
        end;
      end
      else aContext.AddError( 'float expected, no dot, comma or number seen');
    end;
  end;


  function ParseNums( const aContext: TParseContext): TSieveSequence;
  var
    N  : Integer;
    OK : Boolean;
  begin                                                // Nums = num | num , Nums
    OK  := False;
    SetLength( Result, 0);

    while ParseNum( aContext, N)
    do begin
      OK := True;
      SequenceAddNumber( Result, N);

      if   aContext.ParseChar <> ','
      then begin
        aContext.Unparse( 1);                          // We rad past the num .. so backup one
        Break;
      end;
    end;

    if   not OK
    then aContext.AddError( 'number sequence expected');
  end;


  function ParseLiteral( const aContext: TParseContext): TSieveModule;
  var
    aSeq    : TSieveSequence;
    aModulo : Integer;
  begin                                                // Literal = Nums | num
    Result := nil;
    aSeq   := ParseNums( aContext);

    if   aContext.ParseChar = '|'
    then begin
      if   ParseNum( aContext, aModulo)
      then Result := TSieveLiteral.Create( aModulo, aSeq)
      else aContext.AddError( 'modulo expected');
    end
    else aContext.AddError( '| and modulo expected in literal');
  end;


  function ParseModule( const aContext: TParseContext): TSieveModule;
  var
    aModulo  : Integer;
    aResidue : Integer;
  begin                                                // Module = num , num
    Result := nil;

    if   ParseNum( aContext, aModulo)
    then begin
      if   aContext.ParseChar = ','
      then begin
        if   ParseNum( aContext, aResidue)
        then Result := TSieveModule.Create( aModulo, aResidue)
        else aContext.AddError( ', expected in module specification');
      end;
    end;
  end;


  function ParseTerm( const aContext: TParseContext): TSieveModule;
  var
    aTerm   : TSieveModule;
    aName   : string;
    aChance : TSignal;
    anOrder : Integer;
  begin
    Result := nil;

    case aContext.ParseChar of

      '!' :                                            // Term = ! Term
        begin
          aTerm := ParseTerm( aContext);

          if   Assigned( aTerm)
          then Result := aTerm.Complement
          else aContext.AddError( 'Term expected after !');
        end;

      '(' :                                            // Term = ( Sieve )
        begin
          Result := ParseSieve( aContext);

          if   Assigned( Result)
          then begin
            if   aContext.ParseChar <> ')'
            then begin
              FreeAndNil( Result);
              aContext.AddError( '} expected after sieve');
            end;
          end
          else aContext.AddError( 'sieve expected after (');
        end;

      '{' :                                            // Term = { Literal }
        begin
          Result := ParseLiteral( aContext);

          if   Assigned( Result)
          then begin
            if   aContext.ParseChar <> '}'
            then begin
              FreeAndNil( Result);
              aContext.AddError( '} expected after literal');
            end;
          end
          else aContext.AddError( 'literal expected after {');
        end;

      '[' :                                            // Term = [ Module ]
        begin
          Result := ParseModule( aContext);

          if   Assigned( Result)
          then begin
            if   aContext.ParseChar <> ']'
            then begin
              FreeAndNil( Result);
              aContext.AddError( '] expected after module');
            end;
          end
          else aContext.AddError( 'module expected after [');
        end;

      else begin
        aContext.Unparse( 1);
        aName := ParseIdentifier( aContext);

        if   aName <> ''
        then begin
          if      SameText( aName, 'prime')            // Term = 'prime'
          then    Result := TSievePrime.Create
          else if SameText( aName, 'fibonacci')        // Term = 'fibonacci'
          then    Result := TSieveFibonacci.Create
          else if SameText( aName, 'hanoi')            // Term = 'hanoi'
          then    Result := TSieveHanoi.Create
          else if SameText( aName, 'hanoipegs1')       // Term = 'hanoipegs1'
          then    Result := TSieveHanoiPegs1.Create
          else if SameText( aName, 'hanoipegs2')       // Term = 'hanoipegs2'
          then    Result := TSieveHanoiPegs2.Create
          else if SameText( aName, 'hofstadterq')      // Term = 'hofstadterq'
          then    Result := TSieveHofstadterQ.Create
          else if SameText( aName, 'metafib')       // Term = 'metafib' '(' num ')'
          then begin
            if   aContext.ParseChar = '('
            then begin
              if   ParseNum( aContext, anOrder)
              then begin
                if   aContext.ParseChar = ')'
                then Result  := TSieveMetaFib.Create( anOrder)
                else aContext.AddError( ') expected after metafib ( num');
              end
              else aContext.AddError( 'num expected after metafib (');
            end
            else aContext.AddError( '( expected after metafib');
          end
          else if SameText( aName, 'random')           // Term = 'random' '(' float ')'
          then begin
            if   aContext.ParseChar = '('
            then begin
              if   ParseFloat( aContext, aChance)
              then begin
                if   aContext.ParseChar = ')'
                then Result  := TSieveRandom.Create( aChance)
                else aContext.AddError( ') expected after random ( float');
              end
              else aContext.AddError( 'float expected after random (');
            end
            else aContext.AddError( '( expected after random');
          end
          else aContext.AddErrorFmt( 'unknown identifier: ''%s''', [ aName]);
        end
        else aContext.AddError( '!, (, {, [ or identifier expected');
      end;
    end;
  end;


  function ParseSieve( const aContext: TParseContext): TSieveModule;
  var
    t     : char;
    aTerm : TSieveModule;
  begin
    Result := ParseTerm( aContext);                          // Sieve = Term
    t      := aContext.ParseChar;

    while CharInSet( t, [ '+', '*', '^'])
    do begin
      aTerm := ParseTerm( aContext);

      if   Assigned( aTerm)
      then begin
        case t of
          '+'       : Result := Result.Union     ( aTerm);   // Sieve = Term + Term
          '*'       : Result := Result.Intersect ( aTerm);   // Sieve = Term * Term
          '^'       : Result := Result.Difference( aTerm);   // Sieve = Term ^ Term
          Char( -1) : ;                                      // EOT is OK here
          else  aContext.AddError( 'expected +, * or ^');
        end;

        t := aContext.ParseChar;
      end
      else Break;
    end;

    aContext.Unparse( 1);                                    // We read past the num .. so backup one
  end;


  function StrToSieve( var ErrorText: string; const S: string; aPeriod: Integer): TSieveModule;
  var
    aContext : TParseContext;
  begin
    Result := nil;

    if   aPeriod > 0
    then begin
      aContext  := TParseContext.Create( S);
      ErrorText := 'expected a sieve';

      try
        Result    := ParseSieve( aContext);
        ErrorText := aContext.ErrorText;

        if   Assigned( Result)
        then Result.GenerateAll( aPeriod);
      finally
        aContext.DisposeOf;
      end;
    end;
  end;



{ ========
  TSieveModule = class
  private
    FModulo            : Integer;
    FResidue           : Integer;
    FGeneratedSequence : TSieveSequence;
    FGeneratedSignals  : TSieveSignals;
    FIntervals         : TSieveSequence;
    FIntervalSignals   : TSieveSignals;
    FSimplified        : string;
    FValueCount        : Integer;
  public
    property    Modulo            : Integer        read FModulo;
    property    Residue           : Integer        read FResidue;
    property    GeneratedSequence : TSieveSequence read FGeneratedSequence;
    property    GeneratedSignals  : TSieveSignals  read FGeneratedSignals;
    property    Intervals         : TSieveSequence read FIntervals;
    property    IntervalSignals   : TSieveSignals  read FIntervalSignals;
    property    Simplified        : string         read FSimplified;
    property    ValueCount        : Integer        read FValueCount;
  public
}

    constructor TSieveModule.Create( aModulo, aResidue: Integer);
    begin
      Assert(( aModulo > 0) and ( aResidue >= 0));
      FModulo  := aModulo;
      FResidue := aResidue;
    end;


    destructor  TSieveModule.Destroy; // override;
    begin
      inherited;
    end;


    function    TSieveModule.Union( const aSieve: TSieveModule): TSieveModule;
    begin
      Result := TSieveUnion.Create( Self, aSieve);
    end;


    function    TSieveModule.Intersect( const aSieve: TSieveModule): TSieveModule;
    begin
      Result := TSieveIntersection.Create( Self, aSieve);
    end;


    function    TSieveModule.Difference( const aSieve: TSieveModule): TSieveModule;
    begin
      Result := TSieveDifference.Create( Self, aSieve);
    end;


    function    TSieveModule.Complement: TSieveModule;
    begin
      Result := TSieveComplement.Create( Self);
    end;


    function    TSieveModule.HasMember( aValue: Integer): Boolean; // virtual;
    begin
      Result := MathIntMod( aValue, Modulo) = Residue;
    end;


    function    TSieveModule.Dump: string; // virtual;
    begin
      Result := Format( '[%d,%d]', [ Modulo, Residue], AppLocale);
    end;


    function    TSieveModule.Generate( aLimit: Integer): TSieveSequence; // virtual;
    var
      i : Integer;
      p : Integer;
    begin
      Assert( aLimit > 0);

      SetLength( Result, aLimit);
      p := 0;

      for i := 0 to aLimit - 1
      do begin
        if   HasMember( i)
        then begin
          if   p >= Length( Result)
          then Break
          else begin
            Result[ p] := i;
            Inc( p);
          end;
        end;
      end;

      SetLength( Result, p);
    end;


    function    TSieveModule.GenerateSignals( aLimit: Integer): TSieveSignals; // virtual;
    var
      i : Integer;
      p : Integer;
    begin
      Assert( aLimit > 0);

      SetLength( Result, aLimit);
      p := 0;

      for i := 0 to aLimit - 1
      do begin
        if   HasMember( i)
        then begin
          if   p >= Length( Result)
          then Break
          else begin
            Result[ p] := Normalize( i / aLimit);
            Inc( p);
          end;
        end;
      end;

      SetLength( Result, p);
    end;


    function    TSieveModule.GenerateIntervals( aLimit: Integer): TSieveSequence; // virtual;
    var
      anInterval : Integer;
      i          : Integer;
      p          : Integer;
      First      : Integer;
      Last       : Integer;
    begin
      SetLength( Result, aLimit);
      anInterval :=  0;
      p          :=  0;
      First      := -1;
      Last       :=  0;

      for i := 0 to aLimit - 1
      do begin
        if   HasMember( i)
        then begin
          if   First < 0
          then First := i;

          Last := i;

          if   anInterval > 0
          then begin
            Result[ p] := anInterval;
            Inc( p);
          end;

          anInterval := 1;
        end
        else if anInterval > 0
        then Inc( anInterval);
      end;

      if   First >= 0
      then begin
        SetLength( Result, p + 1);
        Result[ p] := aLimit + First - Last;  // Complete the cyclic difference
      end
      else SetLength( Result, 0);             // Empty sieve .. can not calculate any differences, Last = 0, p = 0 then
    end;


    function    TSieveModule.GenerateIntervalSignals( aLimit: Integer): TSieveSignals; // virtual;
    var
      anInterval : Integer;
      i          : Integer;
      p          : Integer;
      First      : Integer;
      Last       : Integer;
      aMax       : TSignal;
    begin
      SetLength( Result, aLimit);
      anInterval := 0;
      p          := 0;
      First      := -1;
      Last       :=  0;
      aMax       := 0.0;

      for i := 0 to aLimit - 1
      do begin
        if   HasMember( i)
        then begin
          if   First < 0
          then First := i;

          Last := i;

          if   anInterval > 0
          then begin
            Result[ p] := anInterval;
            aMax       := Max( aMax, anInterval);
            Inc( p);
          end;

          anInterval := 1;
        end
        else if ( anInterval > 0)
        then Inc( anInterval);
      end;

      if   First >= 0
      then begin
        SetLength( Result, p + 1);
        Result[ p] := aLimit + First - Last;  // Complete the cyclic difference
      end
      else SetLength( Result, 0);             // Empty sieve .. can not calculate any differences, Last = 0, p = 0 then

      if aMax <> 0
      then begin
        for i := 0 to Length( Result) - 1
        do  Result[ i] := Normalize( Result[ i] / aMax);
      end;
    end;


    procedure   TSieveModule.GenerateAll( aLimit: Integer);
    var
      S : TSieveModule;
    begin
      FValueCount := 0;

      if aLimit > 0
      then begin
        FGeneratedSequence := Generate               ( aLimit);
        FGeneratedSignals  := GenerateSignals        ( aLimit);
        FIntervals         := GenerateIntervals      ( aLimit);
        FIntervalSignals   := GenerateIntervalSignals( aLimit);
        FSimplified        := '';
        S                  := SieveSequenceToSieve( FGeneratedSequence, aLimit);

        if Assigned( S)
        then begin
          try
            FSimplified := S.Dump;
          finally
            S.DisposeOf;
          end;
        end
        else FSimplified := '';
      end
      else begin
        SetLength( FGeneratedSequence, 0);
        SetLength( FGeneratedSignals , 0);
        SetLength( FIntervals        , 0);
        SetLength( FIntervalSignals  , 0);
        FSimplified        := '';
      end;

      FValueCount := Length( FGeneratedSequence);
    end;


{ ========
  TSieveLiteral = class( TSieveModule)
  private
    FData : TSieveSet;
  public
}

    constructor TSieveLiteral.Create( aModulo: Integer; const aData: TSieveSequence);
    var
      i : Integer;
      p : Integer;
    begin
      inherited Create( aModulo, 0);
      SetLength( FData, Modulo);
      p := 0;

      for i := 0 to Length( FData) - 1
      do begin
        if   ( p < Length( aData))
        and  ( i = MathIntMod( aData[ p], Modulo))
        then begin
          FData[ i] := True;
          Inc( p);
        end;
      end;
    end;


    function    TSieveLiteral.HasMember( aValue: Integer): Boolean; // override;
    var
      anIndex : Integer;
    begin
      anIndex := MathIntMod( aValue, Modulo);

      if   ( anIndex >= 0)
      and  ( anIndex < Length( FData))
      then Result := FData[ anIndex]
      else Result := False;
    end;


    function    TSieveLiteral.Dump: string; // override;
    var
      i : Integer;
      s : Boolean;
    begin
      s      := False;
      Result := '{';

      for i := 0 to Modulo - 1
      do begin
        if   HasMember( i)
        then begin
          if   s
          then Result := Result + Format( ',%d', [ i], AppLocale)
          else begin
            s         := True;
            Result    := Result + Format(  '%d', [ i], AppLocale);
          end;
        end;
      end;

      Result := Format( '%s|%d}', [ Result, Modulo], AppLocale);
    end;


{ ========
  TSieveExtendable = class( TSieveLiteral)
  public
}

    constructor TSieveExtendable.Create;
    begin
      Inherited Create( 1, [ 0]);
      ExtendTo( 1024);
    end;


    function    TSieveExtendable.HasMember( aValue: Integer): Boolean; // override;
    begin
      if   aValue < 0
      then Result := False
      else begin
        if   aValue >= Length( FData)
        then ExtendTo( aValue * 2);

        Result := FData[ aValue];
      end;
    end;


{ ========
  TSievePrime = class( TSieveExtendable)
  protected
}

    procedure   TSievePrime.ExtendTo( aSize: Integer);
    var
      p : Integer;
    begin
      p := Length( FData);

      if   aSize > p
      then begin
        SetLength( FData, aSize);

        while p < Length( FData)
        do begin
          if   IsPrime( p)
          then FData[ p] := True
          else FData[ p] := False;

          Inc( p);
        end;
      end;
    end;


//  public

    function    TSievePrime.Dump: string; // override;
    begin
      Result := 'prime';
    end;


{ ========
  TSieveFibonacci = class( TSieveExtendable)
  private
}

    procedure   TSieveFibonacci.ExtendTo( aSize: Integer);
    var
      p : Integer;
    begin
      if   Length( FData) < 2
      then FData := [ true, true];        // use {0,1} as a seed

      p := Length( FData);

      if   aSize > p
      then begin
        SetLength( FData, aSize);

        while p < Length( FData)
        do begin
          if   IsFibonacci( p)
          then FData[ p] := True
          else FData[ p] := False;

          Inc( p);
        end;
      end;
    end;


//  public

    function    TSieveFibonacci.Dump: string; // override;
    begin
      Result := 'fibonacci';
    end;


{ ========
  TSievePreExtendable = class( TSieveExtendable)
  private
    FSequence    : TSieveSequence;
    FLastElement : Integer;
  public
    property    Count                     : Integer read GetCount     write SetCount;
    property    LastElement               : Integer read FLastElement;
    property    Element[ anIndex: Integer]: Integer read GetElement;
  protected
}

    function    TSievePreExtendable.GetCount: Integer;
    begin
      Result := Length( FSequence);
    end;


    procedure   TSievePreExtendable.SetCount( aValue: Integer);
    begin
      SetLength( FSequence, aValue);
    end;


    function    TSievePreExtendable.GetElement( anIndex: Integer): Integer;
    begin
      if   ( anIndex >= 0)
      and  ( anIndex < Count)
      then Result := FSequence[ anIndex]
      else Result := 0;;
    end;


    procedure   TSievePreExtendable.AddItem( aValue: Integer);
    begin
      Count                 := Count + 1;
      FSequence[ Count - 1] := aValue;
    end;


    procedure   TSievePreExtendable.AddNextElement;
    begin
      FLastElement := CreateNextElement;
      AddItem( FLastElement);
    end;


    function    TSievePreExtendable.CreateNextElement: Integer; // virtual;
    begin
      Result := 0;
    end;


    procedure   TSievePreExtendable.Initialize; // virtual;
    begin
      FLastElement := -1;
    end;


    procedure   TSievePreExtendable.ExtendTo( aSize: Integer);
    begin
      Initialize;

      if   aSize > Length( FData)
      then begin
        SetLength( FData, aSize); // Which will clear all new elements
        AddNextElement;

        while FLastElement < Length( FData)
        do begin
          FData[ LastElement] := True;
          AddNextElement;
        end;
      end;
    end;


//  public

    function    TSievePreExtendable.Generate( aLimit: Integer): TSieveSequence; // override;
    var
      i : Integer;
      p : Integer;
    begin
      Assert( aLimit > 0);

      if   aLimit > Length( FData)
      then ExtendTo( aLimit);

      Result := FSequence;
      p := 0;

      for i := 0 to Length( Result) - 1
      do  begin
        if   Result[ i] < aLimit
        then Inc( p)
        else Break;
      end;

      SetLength( Result, p);
    end;


    function    TSievePreExtendable.GenerateSignals( aLimit: Integer): TSieveSignals; // override;
    var
      S : TSieveSequence;
      i : Integer;
    begin
      S := Generate( aLimit);

      Assert( aLimit > 0);

      if   aLimit > Length( FData)
      then ExtendTo( aLimit);

      SetLength( Result, Length( S));

      for i := 0 to Length( S) - 1
      do  Result[ i] := Normalize( S[ i] / aLimit);
    end;


    function    TSievePreExtendable.GenerateIntervals( aLimit: Integer): TSieveSequence; // override;
    var
      S : TSieveSequence;
      i : Integer;
    begin
      S := Generate( aLimit);
      SetLength( Result, Length( S));

      for i := 0 to Length( S) - 2
      do Result[ i] := FSequence[ i + 1] - FSequence[ i];

      if   Length( Result) > 0
      then begin
        if   Length( Result) > 1
        then Result[ Length( Result) - 1] := aLimit + S[ 0] - S[ Length( S) - 2]
        else Result[ 0] := 0;
      end;
    end;


    function    TSievePreExtendable.GenerateIntervalSignals( aLimit: Integer): TSieveSignals; // override;
    var
      aSeq : TSieveSequence;
      aMin : TSignal;
      aMax : TSignal;
      i    : Integer;
      V    : TSignal;
    begin
      Assert( aLimit > 0);
      aSeq := GenerateIntervals( aLimit);


      SetLength( Result, Length( aSeq));
      aMin := +1e6;
      aMax := -1e6;

      for i := 0 to Length( aSeq) - 1
      do begin
        V          := aSeq[ i];
        Result[ i] := V;

        if   V < aMin
        then aMin := V;

        if   V > aMax
        then aMax := V;
      end;

      for i := 0 to Length( Result) - 1
      do  begin
        if   aMin = aMax
        then Result[ i] := 0
        else Result[ i] := RangeMap( Result[ i], aMin, aMax, 0.0, 1.0);
      end;
    end;


{ ========
  TSieveHanoi = class( TSievePreExtendable)
  protected
}

    function    TSieveHanoi.CreateNextElement: Integer; // override;
    begin
      Result := FSequence[ Count shr 1] + Count;
    end;


    procedure   TSieveHanoi.Initialize; // override;
    begin
      if   Length( FData) < 2
      then begin
        FData        := [ true, true];        // use {0,1} as a seed
        FSequence    := [ 0, 1];
        FLastElement := 1;
      end;
    end;


//  public

    function    TSieveHanoi.Dump: string; // override;
    begin
      Result := 'hanoi';
    end;


{ ========
  TSieveHanoiPegs = class( TSievePreExtendable)
  var
    FLatestDepth : Integer;
    FOddSwap     : Boolean;
    FSequence    : TSieveSequence;
  private
}

    procedure   TSieveHanoiPegs.AddSeq( aValue: Integer);
    begin
      SetLength( FSequence, Length( FSequence) + 1);
      FSequence[ Length( FSequence) - 1] := aValue;
    end;


//  protected

    procedure   TSieveHanoiPegs.StorePeg( aPeg: Integer);
    begin
      AddSeq( FSequence[ Length( FSequence) - 1] + aPeg);
    end;


    procedure   TSieveHanoiPegs.Tower( aDepth, zero, one, two: Integer);
    begin
      if   aDepth > 0
      then begin
        if   aDepth = 1
        then StorePeg( two)
        else begin
          Tower( aDepth - 1, zero, two, one);
          StorePeg( two);
          Tower( aDepth - 1, one, zero, two);
        end;
      end;
    end;


    procedure   TSieveHanoiPegs.SolveTower( aDepth: Integer);
    begin
      if   aDepth > FLatestDepth
      then begin
        SetLength( FSequence, 0);
        AddSeq( 0);

        if   FOddSwap                   // Every sequence will be a prefix of the following one {1}, [1,2,2],{1,2,2,1,0,1,1} etc.
        then begin
          if   Odd( aDepth)
          then Tower( aDepth, 0, 2, 1)  // When odd  peg 1 is the destination and 2 is the spare
          else Tower( aDepth, 0, 1, 2); // When even peg 2 is the destination and 1 is the spare
        end
        else   Tower( aDepth, 0, 1, 2); // Alternating prefixes {2}, [1,2,2], [2,1,1,2,,0,2,2] etc

        FLatestDepth := aDepth;
      end;
    end;


    procedure   TSieveHanoiPegs.ExtendTo( aSize: Integer); // override;

      function TowerDepth: Integer;
      var
        aDepth : Integer;
      begin
        Result := 0;
        aDepth := 1;

        while ( Result < 32   )
        and   ( aDepth < aSize)
        do begin
          Inc( Result);
          aDepth := ( 1 shl Result) - 1;
        end;
      end;

    var
      aDepth : Integer;
      i      : Integer;
      V      : Integer;
    begin
      if   aSize > Length( FData)
      then begin
        SetLength( FData, aSize);
        aDepth := TowerDepth;
        SolveTower( aDepth);

        for i := 0 to Length( FSequence) - 1
        do begin
          V := FSequence[ i];

          if   V < Length( FData)
          then FData[ V] := True
          else Break;
        end;
      end;
    end;

//  public

    constructor TSieveHanoiPegs.Create( OddSwap: Boolean);
    begin
      FOddSwap := OddSwap;
      inherited Create;
    end;


    function    TSieveHanoiPegs.Dump: string; // override;
    begin
      Result := 'hanoipegs';
    end;


    function    TSieveHanoiPegs.Generate( aLimit: Integer): TSieveSequence; // override;
    var
      i : Integer;
      p : Integer;
    begin
      Assert( aLimit > 0);

      if   aLimit > Length( FData)
      then ExtendTo( aLimit);

      Result := FSequence;
      p := 0;

      for i := 0 to Length( Result) - 1
      do  begin
        if   Result[ i] < aLimit
        then Inc( p)
        else Break;
      end;

      SetLength( Result, p);
    end;


    function    TSieveHanoiPegs.GenerateSignals( aLimit: Integer): TSieveSignals; // override;
    var
      S : TSieveSequence;
      i : Integer;
    begin
      S := Generate( aLimit);

      Assert( aLimit > 0);

      if   aLimit > Length( FData)
      then ExtendTo( aLimit);

      SetLength( Result, Length( S));

      for i := 0 to Length( S) - 1
      do  Result[ i] := Normalize( S[ i] / aLimit);
    end;


    function    TSieveHanoiPegs.GenerateIntervals( aLimit: Integer): TSieveSequence; // override;
    var
      S : TSieveSequence;
      i : Integer;
    begin
      S := Generate( aLimit);
      SetLength( Result, Length( S));

      for i := 0 to Length( S) - 2
      do Result[ i] := FSequence[ i + 1] - FSequence[ i];

      if   Length( Result) > 0
      then begin
        if   Length( Result) > 1
        then Result[ Length( Result) - 1] := aLimit + S[ 0] - S[ Length( S) - 2]
        else Result[ 0] := 0;
      end;
    end;


    function    TSieveHanoiPegs.GenerateIntervalSignals( aLimit: Integer): TSieveSignals; // override;
    var
      aSeq : TSieveSequence;
      aMin : TSignal;
      aMax : TSignal;
      i    : Integer;
      V    : TSignal;
    begin
      Assert( aLimit > 0);
      aSeq := GenerateIntervals( aLimit);


      SetLength( Result, Length( aSeq));
      aMin := +1e6;
      aMax := -1e6;

      for i := 0 to Length( aSeq) - 1
      do begin
        V          := aSeq[ i];
        Result[ i] := V;

        if   V < aMin
        then aMin := V;

        if   V > aMax
        then aMax := V;
      end;

      for i := 0 to Length( Result) - 1
      do begin
        if   aMin = aMax
        then Result[ i] := 0
        else Result[ i] := RangeMap( Result[ i], aMin, aMax, 0.0, 1.0);
      end
    end;


{ ========
  TSieveHanoiPegs1 = class( TSieveHanoiPegs)
  public
}

    constructor TSieveHanoiPegs1.Create;
    begin
      inherited Create( True);
    end;


    function    TSieveHanoiPegs1.Dump: string; // override;
    begin
      Result := 'hanoipegs1';
    end;


{ ========
  TSieveHanoiPegs1 = class( TSieveHanoiPegs)
  public
}

    constructor TSieveHanoiPegs2.Create;
    begin
      inherited Create( False);
    end;


    function    TSieveHanoiPegs2.Dump: string; // override;
    begin
      Result := 'hanoipegs2';
    end;


{ ========
  TSieveHofstadterQ = class( TSievePreExtendable)
  protected
}

    function    TSieveHofstadterQ.CreateNextElement: Integer; // override;
    begin
      Result := Element[ Count - Element[ Count - 1]] + Element[ Count - Element[ Count - 2]];
    end;


    procedure   TSieveHofstadterQ.Initialize; // override;
    begin
      if   Length( FData) < 3
      then begin
        FData        := [ false, true, true];        // use {1,1} as a seed
        FSequence    := [ 1, 1];
        FLastElement := 1;
      end;
    end;


//  public

    function    TSieveHofstadterQ.Dump: string; // override;
    begin
      Result := 'hofstadterq';
    end;


{ ========
  TSieveMetaFib = class( TSievePreExtendable)
  private
    FOrder : Integer;
    FAN    : TSieveSequence;
  public
    property    Order: Integer read FOrder;
  private
}

(*
    We have:

      a[ n] = a[ n - s - a[ n - 1]] + a[ n - s - 1 - a[ n - 2]]; // for n > s + 1
      a[ n] = 1; // for n <= s + 1


    as the starting sequences, s is the 'order' in this. Then the sequence of differences is obtained as :

      d[ n] = a[ n + 1] - a[ n];


    And those sequences have only zeros and ones - or so it seems. Now from d[ n] we construct p[ n] by looking at the positions of the 1's in d[ n] while skipping the zeros, going from left to right first one having index 0. The index of the next '1' then is the next number to be added to p[ n].

    For s = 0 we get as the first few terms:

      a = { 1, 1, 2, 2, 3, 4, 4, 4, 5, 6, 6, 7, 8, 8, 8, 8, 9, 10, 10 ...}
      d = { 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1,  0 ...}
      p = { 1, 3, 4, 7, 8, 10, 11, 15, 16 ...}


    And then the diffrences of p will give the ruler function again, the one this whole thing started with ...

      { 2, 1, 3, 1, 2, 1, 4, 1 ...}
*)

    procedure   TSieveMetaFib.AddAN( aValue: Integer);
    begin
      SetLength( FAN, Length( FAN) + 1);
      FAN[ Length( FAN) - 1] := aValue;
    end;


//  protected

    function    TSieveMetaFib.CreateNextElement: Integer; // override;
    var
      Delta : Integer;
    begin
      repeat
        AddAN( FAN[ Length( FAN) - Order - FAN[ Length( FAN) - 1]] + FAN[ Length( FAN) - Order - 1 - FAN[ Length( FAN) - 2]]); // Next a[ n]
        Delta := FAN[ Length( FAN) - 1] - FAN[ Length( FAN) - 2];                                                              // Next d[ n]

        if   Delta = 1
        then begin
          Result := Length( FAN) - 2; // Next element for p[ n]
          Break;
        end;
      until False;
    end;


    procedure   TSieveMetaFib.Initialize; // override;
    var
      i : Integer;
    begin
      for i := 0 to Order + 1
      do  AddAN( 1);

      FLastElement := 0;
      SetLength( FData, Order + 1);
    end;


//  public

    constructor TSieveMetaFib.Create( anOrder: Integer);
    begin
      FOrder := anOrder;
      inherited Create;
    end;


    function    TSieveMetaFib.Dump: string; // override;
    begin
      Result := Format( 'metafib( %d)', [ Order], AppLocale);
    end;


    function    TSieveMetaFib.DumpAN: string;
    begin
      Result := SieveSequenceToStr( FAN);
    end;


{ ========
  TSieveRandom = class( TSieveExtendable)
  private
    FChance : TSignal;
  protected
}

    procedure   TSieveRandom.ExtendTo( aSize: Integer); // override;
    var
      p : Integer;
    begin
      p := Length( FData);

      if   aSize > p
      then begin
        SetLength( FData, aSize);

        while p < Length( FData)
        do begin
          FData[ p] := Random < FChance;
          Inc( p);
        end;
      end;
    end;


//  public

    constructor TSieveRandom.Create( aChance: TSignal);
    begin
      FChance := aChance;
      inherited Create;
    end;


    function    TSieveRandom.Dump: string; // override;
    begin
      Result := Format( 'random(%g)', [ FChance], AppLocale);
    end;


{ ========
  TSieveUnary = class( TSieveModule)
  private
    FSieve : TSieveModule;
  public
    property    Sieve : TSieveModule read FSieve;
  public
 }

    constructor TSieveUnary.Create( const aSieve: TSieveModule);
    begin
      Assert( Assigned( aSieve));
      FSieve   := aSieve;
      FModulo  := 1;
      FResidue := Sieve.Residue;
    end;


    destructor  TSieveUnary.Destroy; // override;
    begin
      FreeAndNil( FSieve);
      inherited;
    end;


{ ========
  TSieveBinary = class( TSieveModule)
  private
    FLeft  : TSieveModule;
    FRight : TSieveModule;
  public
    property    Left  : TSieveModule read FLeft;
    property    Right : TSieveModule read FRight;
  public
}

    constructor TSieveBinary.Create( const aLeft, aRight: TSieveModule);
    begin
      Assert( Assigned( aLeft) and Assigned( aRight));
      FLeft    := aLeft;
      FRight   := aRight;
      FModulo  := 1;
      FResidue := 0;
    end;


    destructor  TSieveBinary.Destroy; // override;
    begin
      FreeAndNil( FLeft );
      FreeAndNil( FRight);
      inherited;
    end;


{ ========
  TSieveUnion = class( TSieveModule)
  public
}

    function    TSieveUnion.HasMember( aValue: Integer): Boolean; // override;
    begin
      Result := Left.HasMember( aValue) or Right.HasMember( aValue);
    end;


    function    TSieveUnion.Dump: string; // override;
    var
      L : string;
      R : string;
    begin
      if   Assigned( Left)
      then L := Left.Dump
      else L := '-';

      if   Assigned( Right)
      then R := Right.Dump
      else R := '-';

      Result := Format( '%s+%s', [ L, R], AppLocale);
    end;


{ ========
  TSieveIntersection = class( TSieveModule)
  public
}

    function    TSieveIntersection.HasMember( aValue: Integer): Boolean; // override;
    begin
      Result := Left.HasMember( aValue) and Right.HasMember( aValue);
    end;


    function    TSieveIntersection.Dump: string; // override;
    var
      L : string;
      R : string;
    begin
      if   Assigned( Left)
      then L := Left.Dump
      else L := '-';

      if   Assigned( Right)
      then R := Right.Dump
      else R := '-';

      Result := Format( '%s*%s', [ L, R], AppLocale);
    end;


{ ========
  TSieveDifference = class( TSieveModule)
  public
}

    function    TSieveDifference.HasMember( aValue: Integer): Boolean; // override;
    begin
      Result := Left.HasMember( aValue) xor Right.HasMember( aValue);
    end;


    function    TSieveDifference.Dump: string; // override;
    var
      L : string;
      R : string;
    begin
      if   Assigned( Left)
      then L := Left.Dump
      else L := '-';

      if   Assigned( Right)
      then R := Right.Dump
      else R := '-';

      Result := Format( '%s^%s', [ L, R], AppLocale);
    end;


{ ========
  TSieveComplement = class( TSieveUnary)
  public
}

    function    TSieveComplement.HasMember( aValue: Integer): Boolean; // override;
    begin
      Result := not Sieve.HasMember( aValue);
    end;


    function    TSieveComplement.Dump: string; // override;
    begin
      if   Assigned( Sieve)
      then Result := Format( '!%s', [ Sieve.Dump], AppLocale)
      else Result := '!-';

      while Pos( '! !', Result) > 0
      do    Result := StringReplace( Result, '! 1', '!!', []);
    end;


{$ifdef PerformModuleTests}

  procedure TestModule;
  var
    aSequence   : string;
    aDump       : string;
    anIntervals : string;
    ErrorText   : string;
    aModulus    : Integer;
    aResidue    : Integer;
    aSieve      : TSieveModule;
    bSieve      : TSieveModule;
  begin
    bSieve := nil;
    aSieve := TSieveModule.Create( 3, 1);                                   // (3,1)

    try
      aSequence   := SieveSequenceToStr( aSieve.Generate( 12));               // (1,4,7,10)
      anIntervals := SieveSequenceToStr( aSieve.GenerateIntervals( 12));      // {3,3,3.3}
      aDump       := aSieve.Dump;                                             // [3,1]
      aModulus    := aSieve.Modulo;                                           // 3
      aResidue    := aSieve.Residue;                                          // 1

      bSieve      := aSieve.Complement;                                       // (3,1)
      aSequence   := SieveSequenceToStr( bSieve.Generate( 12));               // (0,2,3,5,6,8,9,11)
      anIntervals := SieveSequenceToStr( bSieve.GenerateIntervals( 12));      // {2,1,2,1,2,1,2,1}
      aDump       := bSieve.Dump;                                             // ! [3,1]
      aModulus    := bSieve.Modulo;                                           // 3
      aResidue    := bSieve.Residue;                                          // 1

      aSieve      := TSieveLiteral.Create( 13, [ 1, 3, 8, 9]);                // (13,0)
      aSequence   := SieveSequenceToStr( aSieve.Generate( 12));               // (1,3,8,9)
      anIntervals := SieveSequenceToStr( aSieve.GenerateIntervals( 12));      // {2,5,1,4}
      aDump       := aSieve.Dump;                                             // {1,3,8,9|13}
      aModulus    := aSieve.Modulo;                                           // 13
      aResidue    := aSieve.Residue;                                          //  0

      bSieve      := aSieve.Union( bSieve);                                   // (39,0)
      aSequence   := SieveSequenceToStr( bSieve.Generate( 12));               // (0,1,2,3,5,6,8,9,11)
      anIntervals := SieveSequenceToStr( bSieve.GenerateIntervals( 12));      // {1,1,1,2,1,2,1,2,1}
      aDump       := bSieve.Dump;                                             // ({1,3,8,9|13} + ! [3,1])
      aModulus    := bSieve.Modulo;                                           // 39
      aResidue    := bSieve.Residue;                                          // 0

      aSieve      := TSieveModule.Create( 4, 1);                              // (4,1)
      aSequence   := SieveSequenceToStr( aSieve.Generate( 12));               // (1,5,9)
      anIntervals := SieveSequenceToStr( aSieve.GenerateIntervals( 12));      // {4,4,4}
      aDump       := aSieve.Dump;                                             // [4,1]
      aModulus    := aSieve.Modulo;                                           // 4
      aResidue    := aSieve.Residue;                                          // 1

      bSieve      := aSieve.Intersect( bSieve);                               // (156,0)
      aSequence   := SieveSequenceToStr( bSieve.Generate( 12));               // (1,5,9)
      anIntervals := SieveSequenceToStr( bSieve.GenerateIntervals( 12));      // {4,4,4}
      aDump       := bSieve.Dump;                                             // [4,1] * ({1,3,8,9|13} + ! [3,1])
      aModulus    := bSieve.Modulo;                                           // 156
      aResidue    := bSieve.Residue;                                          //   0

      FreeAndNil( bSieve);

      ErrorText   := '';
      aSieve      := StrToSieve( ErrorText, '([4,1]*({1,3,8,9|13}+![3,1])) + ( prime * fibonacci)');

      if   Assigned( aSieve)
      then begin
        aSequence   := SieveSequenceToStr( aSieve.Generate( 24));             // (1,2,3,5,9,13,17,21)
        anIntervals := SieveSequenceToStr( aSieve.GenerateIntervals( 24));    // {1,1,2,4,4,4,4,4}
        aDump       := aSieve.Dump;                                           // ([4,1] * ({1,3,8,9|13}+![3,1]) + ( prime * fibonacci))
        aModulus    := aSieve.Modulo;                                         // 156
        aResidue    := aSieve.Residue;                                        //   0
        bSieve      := SieveSequenceToSieve( aSieve.Generate( 24), 24);

        if   Assigned( bSieve)
        then begin
          aSequence   := SieveSequenceToStr( bSieve.Generate( 24));           // (1,2,3,5,9,13,17,21)
          anIntervals := SieveSequenceToStr( bSieve.GenerateIntervals( 24));  // {1,1,2,4,4,4,4,4}
          aDump       := bSieve.Dump;                                         // (( [ 4, 1] + [ 11, 2]) + [ 14, 3])
          aModulus    := bSieve.Modulo;                                       // 308
          aResidue    := bSieve.Residue;                                      //   0

          FreeAndNil( bSieve);
        end;

        FreeAndNil( aSieve);
      end;

      ErrorText := '';
      aSieve    := StrToSieve( ErrorText, 'random( 0.33)');

      if   Assigned( aSieve)
      then begin
        aSequence   := SieveSequenceToStr( aSieve.Generate( 24));               // (?)
        anIntervals := SieveSequenceToStr( aSieve.GenerateIntervals( 24));      // {?}
        aDump       := aSieve.Dump;                                             // ( random( 0.33))
        aModulus    := aSieve.Modulo;                                           // 1
        aResidue    := aSieve.Residue;                                          // 0
      end;
    finally
      if   Assigned( bSieve)
      then bSieve.DisposeOf
      else begin
        if   Assigned( aSieve)
        then aSieve.DisposeOf;
      end;
    end;
  end;


  procedure TestHanoi;
  var
    aSieve : TSieveHanoi;
  begin
    aSieve := TSieveHanoi.Create;
    FreeAndNil( aSieve);
  end;


  procedure TestHofstadterQ;
  var
    aSieve : TSieveHofstadterQ;
  begin
    aSieve := TSieveHofstadterQ.Create;
    FreeAndNil( aSieve);
  end;


  procedure TestMetaFibonacci;
  var
    s           : Integer;
    aSieve      : TSieveMetaFib;
    aSequence   : string;
    anIntervals : string;
    anAN        : string;
  begin
    for s := 0 to 9
    do begin
      aSieve := TSieveMetaFib.Create( s);

      try
        anAN        := aSieve.DumpAN;
        aSequence   := SieveSequenceToStr( aSieve.Generate         ( 100));
        anIntervals := SieveSequenceToStr( aSieve.GenerateIntervals( 100));
        Nop;
      finally
        FreeAndNil( aSieve);
      end;
    end;
  end;


  procedure TestHanoiPegs2;
  var
    aSieve      : TSieveHanoiPegs2;
    aSequence   : string;
    anIntervals : string;
  begin
    aSieve := TSieveHanoiPegs2.Create;

    try
      aSequence   := SieveSequenceToStr( aSieve.Generate         ( 100));
      anIntervals := SieveSequenceToStr( aSieve.GenerateIntervals( 100));
    finally
      FreeAndNil( aSieve);
    end;
  end;


{$endif}


initialization



{$ifdef PerformModuleTests}

  TestHanoi;
  TestHofstadterQ;
  TestMetaFibonacci;
  TestHanoiPegs2;
  TestModule;

{$endif}

{$ifndef UseKnobsUtils}

  AppLocale := FormatSettings;
  AppLocale.ThousandSeparator := ',';
  AppLocale.DecimalSeparator  := '.';

{$endif}


end.
